# Copyright (C) 2008, 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Inventory delta serialisation.

See doc/developers/inventory.txt for the description of the format.

In this module the interesting classes are:
 - InventoryDeltaSerializer - object to read/write inventory deltas.
"""

__all__ = ['InventoryDeltaSerializer']

from .. import errors
from ..osutils import basename
from ..revision import NULL_REVISION
from . import inventory

FORMAT_1 = b'bzr inventory delta v1 (bzr 1.14)'


class InventoryDeltaError(errors.BzrError):
    """An error when serializing or deserializing an inventory delta."""

    # Most errors when serializing and deserializing are due to bugs, although
    # damaged input (i.e. a bug in a different process) could cause
    # deserialization errors too.
    internal_error = True

    def __init__(self, format_string, **kwargs):
        # Let each error supply a custom format string and arguments.
        self._fmt = format_string
        super().__init__(**kwargs)


class IncompatibleInventoryDelta(errors.BzrError):
    """The delta could not be deserialised because its contents conflict with
    the allow_versioned_root or allow_tree_references flags of the
    deserializer.
    """
    internal_error = False


def _directory_content(entry):
    """Serialize the content component of entry which is a directory.

    :param entry: An InventoryDirectory.
    """
    return b"dir"


def _file_content(entry):
    """Serialize the content component of entry which is a file.

    :param entry: An InventoryFile.
    """
    if entry.executable:
        exec_bytes = b'Y'
    else:
        exec_bytes = b''
    size_exec_sha = entry.text_size, exec_bytes, entry.text_sha1
    if None in size_exec_sha:
        raise InventoryDeltaError(
            'Missing size or sha for %(fileid)r', fileid=entry.file_id)
    return b"file\x00%d\x00%s\x00%s" % size_exec_sha


def _link_content(entry):
    """Serialize the content component of entry which is a symlink.

    :param entry: An InventoryLink.
    """
    target = entry.symlink_target
    if target is None:
        raise InventoryDeltaError(
            'Missing target for %(fileid)r', fileid=entry.file_id)
    return b"link\x00%s" % target.encode('utf8')


def _reference_content(entry):
    """Serialize the content component of entry which is a tree-reference.

    :param entry: A TreeReference.
    """
    tree_revision = entry.reference_revision
    if tree_revision is None:
        raise InventoryDeltaError(
            'Missing reference revision for %(fileid)r', fileid=entry.file_id)
    return b"tree\x00%s" % tree_revision


def _dir_to_entry(content, name, parent_id, file_id, last_modified,
                  _type=inventory.InventoryDirectory):
    """Convert a dir content record to an InventoryDirectory."""
    result = _type(file_id, name, parent_id)
    result.revision = last_modified
    return result


def _file_to_entry(content, name, parent_id, file_id, last_modified,
                   _type=inventory.InventoryFile):
    """Convert a dir content record to an InventoryFile."""
    result = _type(file_id, name, parent_id)
    result.revision = last_modified
    result.text_size = int(content[1])
    result.text_sha1 = content[3]
    if content[2]:
        result.executable = True
    else:
        result.executable = False
    return result


def _link_to_entry(content, name, parent_id, file_id, last_modified,
                   _type=inventory.InventoryLink):
    """Convert a link content record to an InventoryLink."""
    result = _type(file_id, name, parent_id)
    result.revision = last_modified
    result.symlink_target = content[1].decode('utf8')
    return result


def _tree_to_entry(content, name, parent_id, file_id, last_modified,
                   _type=inventory.TreeReference):
    """Convert a tree content record to a TreeReference."""
    result = _type(file_id, name, parent_id)
    result.revision = last_modified
    result.reference_revision = content[1]
    return result


class InventoryDeltaSerializer:
    """Serialize inventory deltas."""

    def __init__(self, versioned_root, tree_references):
        """Create an InventoryDeltaSerializer.

        :param versioned_root: If True, any root entry that is seen is expected
            to be versioned, and root entries can have any fileid.
        :param tree_references: If True support tree-reference entries.
        """
        self._versioned_root = versioned_root
        self._tree_references = tree_references
        self._entry_to_content = {
            'directory': _directory_content,
            'file': _file_content,
            'symlink': _link_content,
        }
        if tree_references:
            self._entry_to_content['tree-reference'] = _reference_content

    def delta_to_lines(self, old_name, new_name, delta_to_new):
        """Return a line sequence for delta_to_new.

        Both the versioned_root and tree_references flags must be set via
        require_flags before calling this.

        :param old_name: A UTF8 revision id for the old inventory.  May be
            NULL_REVISION if there is no older inventory and delta_to_new
            includes the entire inventory contents.
        :param new_name: The version name of the inventory we create with this
            delta.
        :param delta_to_new: An inventory delta such as Inventory.apply_delta
            takes.
        :return: The serialized delta as lines.
        """
        if not isinstance(old_name, bytes):
            raise TypeError('old_name should be str, got {!r}'.format(old_name))
        if not isinstance(new_name, bytes):
            raise TypeError('new_name should be str, got {!r}'.format(new_name))
        lines = [b'', b'', b'', b'', b'']
        to_line = self._delta_item_to_line
        for delta_item in delta_to_new:
            line = to_line(delta_item, new_name)
            # GZ 2017-06-09: Not really worth asserting this here
            if line.__class__ != bytes:
                raise InventoryDeltaError(
                    'to_line gave non-bytes output %(line)r', line=lines[-1])
            lines.append(line)
        lines.sort()
        lines[0] = b"format: %s\n" % FORMAT_1
        lines[1] = b"parent: %s\n" % old_name
        lines[2] = b"version: %s\n" % new_name
        lines[3] = b"versioned_root: %s\n" % self._serialize_bool(
            self._versioned_root)
        lines[4] = b"tree_references: %s\n" % self._serialize_bool(
            self._tree_references)
        return lines

    def _serialize_bool(self, value):
        if value:
            return b"true"
        else:
            return b"false"

    def _delta_item_to_line(self, delta_item, new_version):
        """Convert delta_item to a line."""
        oldpath, newpath, file_id, entry = delta_item
        if newpath is None:
            # delete
            oldpath_utf8 = b'/' + oldpath.encode('utf8')
            newpath_utf8 = b'None'
            parent_id = b''
            last_modified = NULL_REVISION
            content = b'deleted\x00\x00'
        else:
            if oldpath is None:
                oldpath_utf8 = b'None'
            else:
                oldpath_utf8 = b'/' + oldpath.encode('utf8')
            if newpath == '/':
                raise AssertionError(
                    "Bad inventory delta: '/' is not a valid newpath "
                    "(should be '') in delta item %r" % (delta_item,))
            # TODO: Test real-world utf8 cache hit rate. It may be a win.
            newpath_utf8 = b'/' + newpath.encode('utf8')
            # Serialize None as ''
            parent_id = entry.parent_id or b''
            # Serialize unknown revisions as NULL_REVISION
            last_modified = entry.revision
            # special cases for /
            if newpath_utf8 == b'/' and not self._versioned_root:
                # This is an entry for the root, this inventory does not
                # support versioned roots.  So this must be an unversioned
                # root, i.e. last_modified == new revision.  Otherwise, this
                # delta is invalid.
                # Note: the non-rich-root repositories *can* have roots with
                # file-ids other than TREE_ROOT, e.g. repo formats that use the
                # xml5 serializer.
                if last_modified != new_version:
                    raise InventoryDeltaError(
                        'Version present for / in %(fileid)r'
                        ' (%(last)r != %(new)r)',
                        fileid=file_id, last=last_modified, new=new_version)
            if last_modified is None:
                raise InventoryDeltaError(
                    "no version for fileid %(fileid)r", fileid=file_id)
            content = self._entry_to_content[entry.kind](entry)
        return (b"%s\x00%s\x00%s\x00%s\x00%s\x00%s\n" %
                (oldpath_utf8, newpath_utf8, file_id, parent_id, last_modified,
                 content))


class InventoryDeltaDeserializer:
    """Deserialize inventory deltas."""

    def __init__(self, allow_versioned_root=True, allow_tree_references=True):
        """Create an InventoryDeltaDeserializer.

        :param versioned_root: If True, any root entry that is seen is expected
            to be versioned, and root entries can have any fileid.
        :param tree_references: If True support tree-reference entries.
        """
        self._allow_versioned_root = allow_versioned_root
        self._allow_tree_references = allow_tree_references

    def _deserialize_bool(self, value):
        if value == b"true":
            return True
        elif value == b"false":
            return False
        else:
            raise InventoryDeltaError("value %(val)r is not a bool", val=value)

    def parse_text_bytes(self, lines):
        """Parse the text bytes of a serialized inventory delta.

        If versioned_root and/or tree_references flags were set via
        require_flags, then the parsed flags must match or a BzrError will be
        raised.

        :param lines: The lines to parse. This can be obtained by calling
            delta_to_lines.
        :return: (parent_id, new_id, versioned_root, tree_references,
            inventory_delta)
        """
        if not lines:
            raise InventoryDeltaError(
                'inventory delta is empty')
        if not lines[-1].endswith(b'\n'):
            raise InventoryDeltaError(
                'last line not empty: %(line)r', line=lines[-1])
        lines = [line.rstrip(b'\n') for line in lines]  # discard the last empty line
        if not lines or lines[0] != b'format: %s' % FORMAT_1:
            raise InventoryDeltaError(
                'unknown format %(line)r', line=lines[0:1])
        if len(lines) < 2 or not lines[1].startswith(b'parent: '):
            raise InventoryDeltaError('missing parent: marker')
        delta_parent_id = lines[1][8:]
        if len(lines) < 3 or not lines[2].startswith(b'version: '):
            raise InventoryDeltaError('missing version: marker')
        delta_version_id = lines[2][9:]
        if len(lines) < 4 or not lines[3].startswith(b'versioned_root: '):
            raise InventoryDeltaError('missing versioned_root: marker')
        delta_versioned_root = self._deserialize_bool(lines[3][16:])
        if len(lines) < 5 or not lines[4].startswith(b'tree_references: '):
            raise InventoryDeltaError('missing tree_references: marker')
        delta_tree_references = self._deserialize_bool(lines[4][17:])
        if (not self._allow_versioned_root and delta_versioned_root):
            raise IncompatibleInventoryDelta("versioned_root not allowed")
        result = []
        seen_ids = set()
        line_iter = iter(lines)
        for i in range(5):
            next(line_iter)
        for line in line_iter:
            (oldpath_utf8, newpath_utf8, file_id, parent_id, last_modified,
                content) = line.split(b'\x00', 5)
            parent_id = parent_id or None
            if file_id in seen_ids:
                raise InventoryDeltaError(
                    "duplicate file id %(fileid)r", fileid=file_id)
            seen_ids.add(file_id)
            if (newpath_utf8 == b'/' and not delta_versioned_root and
                    last_modified != delta_version_id):
                # Delta claims to be not have a versioned root, yet here's
                # a root entry with a non-default version.
                raise InventoryDeltaError(
                    "Versioned root found: %(line)r", line=line)
            elif newpath_utf8 != b'None' and last_modified[-1:] == b':':
                # Deletes have a last_modified of null:, but otherwise special
                # revision ids should not occur.
                raise InventoryDeltaError(
                    'special revisionid found: %(line)r', line=line)
            if content.startswith(b'tree\x00'):
                if delta_tree_references is False:
                    raise InventoryDeltaError(
                        "Tree reference found (but header said "
                        "tree_references: false): %(line)r", line=line)
                elif not self._allow_tree_references:
                    raise IncompatibleInventoryDelta(
                        "Tree reference not allowed")
            if oldpath_utf8 == b'None':
                oldpath = None
            elif oldpath_utf8[:1] != b'/':
                raise InventoryDeltaError(
                    "oldpath invalid (does not start with /): %(path)r",
                    path=oldpath_utf8)
            else:
                oldpath_utf8 = oldpath_utf8[1:]
                oldpath = oldpath_utf8.decode('utf8')
            if newpath_utf8 == b'None':
                newpath = None
            elif newpath_utf8[:1] != b'/':
                raise InventoryDeltaError(
                    "newpath invalid (does not start with /): %(path)r",
                    path=newpath_utf8)
            else:
                # Trim leading slash
                newpath_utf8 = newpath_utf8[1:]
                newpath = newpath_utf8.decode('utf8')
            content_tuple = tuple(content.split(b'\x00'))
            if content_tuple[0] == b'deleted':
                entry = None
            else:
                entry = _parse_entry(
                    newpath, file_id, parent_id, last_modified, content_tuple)
            delta_item = (oldpath, newpath, file_id, entry)
            result.append(delta_item)
        return (delta_parent_id, delta_version_id, delta_versioned_root,
                delta_tree_references, result)


def _parse_entry(path, file_id, parent_id, last_modified, content):
    entry_factory = {
        b'dir': _dir_to_entry,
        b'file': _file_to_entry,
        b'link': _link_to_entry,
        b'tree': _tree_to_entry,
    }
    kind = content[0]
    if path.startswith('/'):
        raise AssertionError
    name = basename(path)
    return entry_factory[content[0]](
        content, name, parent_id, file_id, last_modified)



# Copyright (C) 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import sys
import time

from breezy import errors
from breezy import revision as _mod_revision
from breezy import tests, transform
from breezy.bzr import inventory, remote
from breezy.bzr.tests.per_repository_vf import (
    TestCaseWithRepository, all_repository_vf_format_scenarios)
from breezy.tests.scenarios import load_tests_apply_scenarios

load_tests = load_tests_apply_scenarios


class FileIdInvolvedWGhosts(TestCaseWithRepository):

    scenarios = all_repository_vf_format_scenarios()

    def create_branch_with_ghost_text(self):
        builder = self.make_branch_builder('ghost')
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', None)),
            ('add', ('a', b'a-file-id', 'file', b'some content\n'))],
            revision_id=b'A-id')
        b = builder.get_branch()
        old_rt = b.repository.revision_tree(b'A-id')
        new_inv = inventory.mutable_inventory_from_tree(old_rt)
        new_inv.revision_id = b'B-id'
        new_inv.get_entry(b'a-file-id').revision = b'ghost-id'
        new_rev = _mod_revision.Revision(b'B-id',
                                         timestamp=time.time(),
                                         timezone=0,
                                         message='Committing against a ghost',
                                         committer='Joe Foo <joe@foo.com>',
                                         properties={},
                                         parent_ids=(b'A-id', b'ghost-id'),
                                         )
        b.lock_write()
        self.addCleanup(b.unlock)
        b.repository.start_write_group()
        b.repository.add_revision(b'B-id', new_rev, new_inv)
        self.disable_commit_write_group_paranoia(b.repository)
        b.repository.commit_write_group()
        return b

    def disable_commit_write_group_paranoia(self, repo):
        if isinstance(repo, remote.RemoteRepository):
            # We can't easily disable the checks in a remote repo.
            repo.abort_write_group()
            raise tests.TestSkipped(
                "repository format does not support storing revisions with "
                "missing texts.")
        pack_coll = getattr(repo, '_pack_collection', None)
        if pack_coll is not None:
            # Monkey-patch the pack collection instance to allow storing
            # incomplete revisions.
            pack_coll._check_new_inventories = lambda: []

    def test_file_ids_include_ghosts(self):
        b = self.create_branch_with_ghost_text()
        repo = b.repository
        self.assertEqual(
            {b'a-file-id': {b'ghost-id'}},
            repo.fileids_altered_by_revision_ids([b'B-id']))

    def test_file_ids_uses_fallbacks(self):
        builder = self.make_branch_builder('source',
                                           format=self.bzrdir_format)
        repo = builder.get_branch().repository
        if not repo._format.supports_external_lookups:
            raise tests.TestNotApplicable('format does not support stacking')
        builder.start_series()
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', None)),
            ('add', ('file', b'file-id', 'file', b'contents\n'))],
            revision_id=b'A-id')
        builder.build_snapshot([b'A-id'], [
            ('modify', ('file', b'new-content\n'))],
            revision_id=b'B-id')
        builder.build_snapshot([b'B-id'], [
            ('modify', ('file', b'yet more content\n'))],
            revision_id=b'C-id')
        builder.finish_series()
        source_b = builder.get_branch()
        source_b.lock_read()
        self.addCleanup(source_b.unlock)
        base = self.make_branch('base')
        base.pull(source_b, stop_revision=b'B-id')
        stacked = self.make_branch('stacked')
        stacked.set_stacked_on_url('../base')
        stacked.pull(source_b, stop_revision=b'C-id')

        stacked.lock_read()
        self.addCleanup(stacked.unlock)
        repo = stacked.repository
        keys = {b'file-id': {b'A-id'}}
        if stacked.repository.supports_rich_root():
            keys[b'root-id'] = {b'A-id'}
        self.assertEqual(keys, repo.fileids_altered_by_revision_ids([b'A-id']))


class FileIdInvolvedBase(TestCaseWithRepository):

    def touch(self, tree, filename):
        # use the trees transport to not depend on the tree's location or type.
        tree.controldir.root_transport.append_bytes(
            filename, b"appended line\n")

    def compare_tree_fileids(self, branch, old_rev, new_rev):
        old_tree = self.branch.repository.revision_tree(old_rev)
        new_tree = self.branch.repository.revision_tree(new_rev)
        delta = new_tree.changes_from(old_tree)

        l2 = [change.file_id for change in delta.added] + \
             [change.file_id for change in delta.renamed] + \
             [change.file_id for change in delta.modified] + \
             [change.file_id for change in delta.copied]
        return set(l2)


class TestFileIdInvolved(FileIdInvolvedBase):

    scenarios = all_repository_vf_format_scenarios()

    def setUp(self):
        super().setUp()
        # create three branches, and merge it
        #
        #          ,-->J------>K                (branch2)
        #         /             \
        #  A --->B --->C---->D-->G              (main)
        #  \          /     /
        #   '--->E---+---->F                    (branch1)

        # A changes:
        # B changes: 'a-file-id-2006-01-01-abcd'
        # C changes:  Nothing (perfect merge)
        # D changes: 'b-file-id-2006-01-01-defg'
        # E changes: 'file-d'
        # F changes: 'file-d'
        # G changes: 'b-file-id-2006-01-01-defg'
        # J changes: 'b-file-id-2006-01-01-defg'
        # K changes: 'c-funky<file-id>quiji%bo'

        main_wt = self.make_branch_and_tree('main')
        main_branch = main_wt.branch
        self.build_tree(["main/a", "main/b", "main/c"])

        main_wt.add(
            ['a', 'b', 'c'],
            ids=[b'a-file-id-2006-01-01-abcd',
                 b'b-file-id-2006-01-01-defg',
                 b'c-funky<file-id>quiji%bo'])
        try:
            main_wt.commit("Commit one", rev_id=b"rev-A")
        except errors.IllegalPath:
            # TODO: jam 20060701 Consider raising a different exception
            #       newer formats do support this, and nothin can done to
            #       correct this test - its not a bug.
            if sys.platform == 'win32':
                raise tests.TestSkipped('Old repository formats do not'
                                        ' support file ids with <> on win32')
            # This is not a known error condition
            raise

        # -------- end A -----------

        bt1 = self.make_branch_and_tree('branch1')
        bt1.pull(main_branch)
        b1 = bt1.branch
        self.build_tree(["branch1/d"])
        bt1.add(['d'], ids=[b'file-d'])
        bt1.commit("branch1, Commit one", rev_id=b"rev-E")

        # -------- end E -----------

        self.touch(main_wt, "a")
        main_wt.commit("Commit two", rev_id=b"rev-B")

        # -------- end B -----------

        bt2 = self.make_branch_and_tree('branch2')
        bt2.pull(main_branch)
        branch2_branch = bt2.branch
        set_executability(bt2, 'b', True)
        bt2.commit("branch2, Commit one", rev_id=b"rev-J")

        # -------- end J -----------

        main_wt.merge_from_branch(b1)
        main_wt.commit("merge branch1, rev-11", rev_id=b"rev-C")

        # -------- end C -----------

        bt1.rename_one("d", "e")
        bt1.commit("branch1, commit two", rev_id=b"rev-F")

        # -------- end F -----------

        self.touch(bt2, "c")
        bt2.commit("branch2, commit two", rev_id=b"rev-K")

        # -------- end K -----------

        main_wt.merge_from_branch(b1)
        self.touch(main_wt, "b")
        # D gets some funky characters to make sure the unescaping works
        main_wt.commit("merge branch1, rev-12", rev_id=b"rev-<D>")

        # end D

        main_wt.merge_from_branch(branch2_branch)
        main_wt.commit("merge branch1, rev-22", rev_id=b"rev-G")

        # end G
        self.branch = main_branch

    def test_fileids_altered_between_two_revs(self):
        self.branch.lock_read()
        self.addCleanup(self.branch.unlock)
        self.branch.repository.fileids_altered_by_revision_ids(
            [b"rev-J", b"rev-K"])
        self.assertEqual(
            {b'b-file-id-2006-01-01-defg': {b'rev-J'},
             b'c-funky<file-id>quiji%bo': {b'rev-K'}
             },
            self.branch.repository.fileids_altered_by_revision_ids([b"rev-J", b"rev-K"]))

        self.assertEqual(
            {b'b-file-id-2006-01-01-defg': {b'rev-<D>'},
             b'file-d': {b'rev-F'},
             },
            self.branch.repository.fileids_altered_by_revision_ids([b'rev-<D>', b'rev-F']))

        self.assertEqual(
            {
                b'b-file-id-2006-01-01-defg': {b'rev-<D>', b'rev-G', b'rev-J'},
                b'c-funky<file-id>quiji%bo': {b'rev-K'},
                b'file-d': {b'rev-F'},
                },
            self.branch.repository.fileids_altered_by_revision_ids(
                [b'rev-<D>', b'rev-G', b'rev-F', b'rev-K', b'rev-J']))

        self.assertEqual(
            {b'a-file-id-2006-01-01-abcd': {b'rev-B'},
             b'b-file-id-2006-01-01-defg': {b'rev-<D>', b'rev-G', b'rev-J'},
             b'c-funky<file-id>quiji%bo': {b'rev-K'},
             b'file-d': {b'rev-F'},
             },
            self.branch.repository.fileids_altered_by_revision_ids(
                [b'rev-G', b'rev-F', b'rev-C', b'rev-B', b'rev-<D>', b'rev-K', b'rev-J']))

    def fileids_altered_by_revision_ids(self, revision_ids):
        """This is a wrapper to strip TREE_ROOT if it occurs"""
        repo = self.branch.repository
        root_id = self.branch.basis_tree().path2id('')
        result = repo.fileids_altered_by_revision_ids(revision_ids)
        if root_id in result:
            del result[root_id]
        return result

    def test_fileids_altered_by_revision_ids(self):
        self.branch.lock_read()
        self.addCleanup(self.branch.unlock)
        self.assertEqual(
            {b'a-file-id-2006-01-01-abcd': {b'rev-A'},
             b'b-file-id-2006-01-01-defg': {b'rev-A'},
             b'c-funky<file-id>quiji%bo': {b'rev-A'},
             },
            self.fileids_altered_by_revision_ids([b"rev-A"]))
        self.assertEqual(
            {b'a-file-id-2006-01-01-abcd': {b'rev-B'}
             },
            self.branch.repository.fileids_altered_by_revision_ids([b"rev-B"]))
        self.assertEqual(
            {b'b-file-id-2006-01-01-defg': {b'rev-<D>'}
             },
            self.branch.repository.fileids_altered_by_revision_ids([b"rev-<D>"]))

    def test_fileids_involved_full_compare(self):
        # this tests that the result of each fileid_involved calculation
        # along a revision history selects only the fileids selected by
        # comparing the trees - no less, and no more. This is correct
        # because in our sample data we do not revert any file ids along
        # the revision history.
        self.branch.lock_read()
        self.addCleanup(self.branch.unlock)
        pp = []
        graph = self.branch.repository.get_graph()
        history = list(graph.iter_lefthand_ancestry(self.branch.last_revision(),
                                                    [_mod_revision.NULL_REVISION]))
        history.reverse()

        if len(history) < 2:
            return

        for start in range(0, len(history) - 1):
            start_id = history[start]
            for end in range(start + 1, len(history)):
                end_id = history[end]
                unique_revs = graph.find_unique_ancestors(end_id, [start_id])
                l1 = self.branch.repository.fileids_altered_by_revision_ids(
                    unique_revs)
                l1 = set(l1.keys())
                l2 = self.compare_tree_fileids(self.branch, start_id, end_id)
                self.assertEqual(l1, l2)


class TestFileIdInvolvedNonAscii(FileIdInvolvedBase):

    scenarios = all_repository_vf_format_scenarios()

    def test_utf8_file_ids_and_revision_ids(self):
        main_wt = self.make_branch_and_tree('main')
        main_branch = main_wt.branch
        self.build_tree(["main/a"])

        file_id = 'a-f\xedle-id'.encode()
        main_wt.add(['a'], ids=[file_id])
        revision_id = 'r\xe9v-a'.encode()
        try:
            main_wt.commit('a', rev_id=revision_id)
        except errors.NonAsciiRevisionId:
            raise tests.TestSkipped('non-ascii revision ids not supported by %s'
                                    % self.repository_format)

        repo = main_wt.branch.repository
        repo.lock_read()
        self.addCleanup(repo.unlock)
        file_ids = repo.fileids_altered_by_revision_ids([revision_id])
        root_id = main_wt.basis_tree().path2id('')
        if root_id in file_ids:
            self.assertEqual({file_id: {revision_id},
                              root_id: {revision_id}
                              }, file_ids)
        else:
            self.assertEqual({file_id: {revision_id}}, file_ids)


class TestFileIdInvolvedSuperset(FileIdInvolvedBase):

    scenarios = all_repository_vf_format_scenarios()

    def setUp(self):
        super().setUp()

        self.branch = None
        main_wt = self.make_branch_and_tree('main')
        main_branch = main_wt.branch
        self.build_tree(["main/a", "main/b", "main/c"])

        main_wt.add(
            ['a', 'b', 'c'],
            ids=[b'a-file-id-2006-01-01-abcd',
                 b'b-file-id-2006-01-01-defg',
                 b'c-funky<file-id>quiji\'"%bo'])
        try:
            main_wt.commit("Commit one", rev_id=b"rev-A")
        except errors.IllegalPath:
            # TODO: jam 20060701 Consider raising a different exception
            #       newer formats do support this, and nothin can done to
            #       correct this test - its not a bug.
            if sys.platform == 'win32':
                raise tests.TestSkipped('Old repository formats do not'
                                        ' support file ids with <> on win32')
            # This is not a known error condition
            raise

        branch2_wt = self.make_branch_and_tree('branch2')
        branch2_wt.pull(main_branch)
        branch2_bzrdir = branch2_wt.controldir
        branch2_branch = branch2_bzrdir.open_branch()
        set_executability(branch2_wt, 'b', True)
        branch2_wt.commit("branch2, Commit one", rev_id=b"rev-J")

        main_wt.merge_from_branch(branch2_branch)
        set_executability(main_wt, 'b', False)
        main_wt.commit("merge branch1, rev-22", rev_id=b"rev-G")

        # end G
        self.branch = main_branch

    def test_fileid_involved_full_compare2(self):
        # this tests that fileids_altered_by_revision_ids returns
        # more information than compare_tree can, because it
        # sees each change rather than the aggregate delta.
        self.branch.lock_read()
        self.addCleanup(self.branch.unlock)
        graph = self.branch.repository.get_graph()
        history = list(graph.iter_lefthand_ancestry(self.branch.last_revision(),
                                                    [_mod_revision.NULL_REVISION]))
        history.reverse()
        old_rev = history[0]
        new_rev = history[1]
        unique_revs = graph.find_unique_ancestors(new_rev, [old_rev])

        l1 = self.branch.repository.fileids_altered_by_revision_ids(
            unique_revs)
        l1 = set(l1.keys())

        l2 = self.compare_tree_fileids(self.branch, old_rev, new_rev)
        self.assertNotEqual(l2, l1)
        self.assertSubset(l2, l1)


def set_executability(wt, path, executable=True):
    """Set the executable bit for the file at path in the working tree

    os.chmod() doesn't work on windows. But TreeTransform can mark or
    unmark a file as executable.
    """
    with wt.transform() as tt:
        tt.set_executability(executable, tt.trans_id_tree_path(path))
        tt.apply()



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""bzr-commitfromnews - make commit messages from the changes in a NEWS file.

commitfromnews is enabled by default when installed.

To use, set the ``commit.template_from_files`` setting to a path and
just do a commit where the NEWS file for your project has a new section
added without providing a message to commit.

E.g.::
  $ echo "commit.template_from_files = NEWS" >> .bzr/branch/branch.conf
  $ echo "\n* new thing\n" >> NEWS
  $ bzr commit
  # editor pops open to let you tweak the message, and it starts with
    "* new thing" as the message to edit.

commitfromnews attempts to create a sensible default commit message by
including sections from a NEWS or ChangeLog file.
"""

from ... import version_info  # noqa: F401
from ... import hooks
from ...config import ListOption, option_registry

option_registry.register(
    ListOption('commit.template_from_files', default=[], help="""\
List of fnmatch(2)-style shell file patterns to use when creating commit
templates.
"""))


def commit_template(commit, message):
    """Create a commit message for commit based on changes in the tree."""
    config_stack = commit.work_tree.get_config_stack()
    filespec = config_stack.get('commit.template_from_files')
    if filespec:
        from .committemplate import CommitTemplate
        template = CommitTemplate(commit, message, filespec)
        return template.make()
    return message


def load_tests(loader, basic_tests, pattern):
    testmod_names = [
        'tests',
        ]
    basic_tests.addTest(loader.loadTestsFromModuleNames(
        ["{}.{}".format(__name__, tmn) for tmn in testmod_names]))
    return basic_tests


_registered = False


def register():
    """Register the plugin."""
    global _registered
    # Does not check registered because only tests call this, and they are
    # isolated.
    _registered = True
    hooks.install_lazy_named_hook(
        'breezy.msgeditor', 'hooks',
        'commit_message_template',
        commit_template, 'commitfromnews template')


register()



# Copyright (C) 2005-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import sys

from . import delta as _mod_delta
from . import errors as errors
from . import hooks as _mod_hooks
from . import log, osutils
from . import revision as _mod_revision
from . import tsort
from .trace import mutter, warning
from .workingtree import ShelvingUnsupported

# TODO: when showing single-line logs, truncate to the width of the terminal
# if known, but only if really going to the terminal (not into a file)


def report_changes(to_file, old, new, specific_files,
                   show_short_reporter, show_long_callback,
                   short=False, want_unchanged=False,
                   want_unversioned=False, show_ids=False, classify=True):
    """Display summary of changes.

    This compares two trees with regards to a list of files, and delegates
    the display to underlying elements.

    For short output, it creates an iterator on all changes, and lets a given
    reporter display these changes.

    For stantard output, it creates a delta of the changes, and forwards it
    to a callback

    :param to_file: If set, write to this file (default stdout.)
    :param old: Start tree for the comparison
    :param end: End tree for the comparison
    :param specific_files: If set, a list of filenames whose status should be
        shown.  It is an error to give a filename that is not in the working
        tree, or in the working inventory or in the basis inventory.
    :param show_short_reporter: Reporter in charge of display for short output
    :param show_long_callback: Callback in charge of display for normal output
    :param short: If True, gives short SVN-style status lines.
    :param want_unchanged: Deprecated parameter. If set, includes unchanged
        files.
    :param show_ids: If set, includes each file's id.
    :param want_unversioned: If False, only shows versioned files.
    :param classify: Add special symbols to indicate file kind.
    """

    if short:
        changes = new.iter_changes(old, want_unchanged, specific_files,
                                   require_versioned=False, want_unversioned=want_unversioned)
        _mod_delta.report_changes(changes, show_short_reporter)
    else:
        delta = new.changes_from(old, want_unchanged=want_unchanged,
                                 specific_files=specific_files,
                                 want_unversioned=want_unversioned)
        # filter out unknown files. We may want a tree method for
        # this
        delta.unversioned = [change for change in delta.unversioned
                             if not new.is_ignored(change.path[1])]
        show_long_callback(to_file, delta,
                           show_ids=show_ids,
                           show_unchanged=want_unchanged,
                           classify=classify)


def show_tree_status(wt,
                     specific_files=None,
                     show_ids=False,
                     to_file=None,
                     show_pending=True,
                     revision=None,
                     short=False,
                     verbose=False,
                     versioned=False,
                     classify=True,
                     show_long_callback=_mod_delta.report_delta):
    """Display summary of changes.

    By default this compares the working tree to a previous revision.
    If the revision argument is given, summarizes changes between the
    working tree and another, or between two revisions.

    The result is written out as Unicode and to_file should be able
    to encode that.

    If showing the status of a working tree, extra information is included
    about unknown files, conflicts, and pending merges.

    :param specific_files: If set, a list of filenames whose status should be
        shown.  It is an error to give a filename that is not in the working
        tree, or in the working inventory or in the basis inventory.
    :param show_ids: If set, includes each file's id.
    :param to_file: If set, write to this file (default stdout.)
    :param show_pending: If set, write pending merges.
    :param revision: If None, compare latest revision with working tree
        If not None, it must be a RevisionSpec list.
        If one revision, compare with working tree.
        If two revisions, show status between first and second.
    :param short: If True, gives short SVN-style status lines.
    :param verbose: If True, show all merged revisions, not just
        the merge tips
    :param versioned: If True, only shows versioned files.
    :param classify: Add special symbols to indicate file kind.
    :param show_long_callback: A callback: message = show_long_callback(to_file, delta,
        show_ids, show_unchanged, indent, filter), only used with the long output
    """
    if to_file is None:
        to_file = sys.stdout

    with wt.lock_read():
        new_is_working_tree = True
        if revision is None:
            if wt.last_revision() != wt.branch.last_revision():
                warning("working tree is out of date, run 'brz update'")
            new = wt
            old = new.basis_tree()
        elif len(revision) > 0:
            try:
                old = revision[0].as_tree(wt.branch)
            except errors.NoSuchRevision as e:
                raise errors.CommandError(str(e))
            if (len(revision) > 1) and (revision[1].spec is not None):
                try:
                    new = revision[1].as_tree(wt.branch)
                    new_is_working_tree = False
                except errors.NoSuchRevision as e:
                    raise errors.CommandError(str(e))
            else:
                new = wt
        with old.lock_read(), new.lock_read():
            for hook in hooks['pre_status']:
                hook(StatusHookParams(
                    old, new, to_file, versioned, show_ids, short, verbose,
                    specific_files=specific_files))

            specific_files, nonexistents \
                = _filter_nonexistent(specific_files, old, new)
            want_unversioned = not versioned

            # Reporter used for short outputs
            reporter = _mod_delta._ChangeReporter(
                output_file=to_file, unversioned_filter=new.is_ignored,
                classify=classify)
            report_changes(to_file, old, new, specific_files,
                           reporter, show_long_callback,
                           short=short, want_unversioned=want_unversioned,
                           show_ids=show_ids, classify=classify)

            # show the ignored files among specific files (i.e. show the files
            # identified from input that we choose to ignore).
            if specific_files is not None:
                # Ignored files is sorted because specific_files is already sorted
                ignored_files = [specific for specific in
                                 specific_files if new.is_ignored(specific)]
                if len(ignored_files) > 0 and not short:
                    to_file.write("ignored:\n")
                    prefix = ' '
                else:
                    prefix = 'I  '
                for ignored_file in ignored_files:
                    to_file.write("{} {}\n".format(prefix, ignored_file))

            # show the new conflicts only for now. XXX: get them from the
            # delta.
            conflicts = new.conflicts()
            if specific_files is not None:
                conflicts = conflicts.select_conflicts(
                    new, specific_files, ignore_misses=True, recurse=True)[1]
            if len(conflicts) > 0 and not short:
                to_file.write("conflicts:\n")
            for conflict in conflicts:
                if short:
                    prefix = 'C  '
                else:
                    prefix = ' '
                to_file.write("{} {}\n".format(prefix, conflict.describe()))
            # Show files that were requested but don't exist (and are
            # not versioned).  We don't involve delta in this; these
            # paths are really the province of just the status
            # command, since they have more to do with how it was
            # invoked than with the tree it's operating on.
            if nonexistents and not short:
                to_file.write("nonexistent:\n")
            for nonexistent in nonexistents:
                # We could calculate prefix outside the loop but, given
                # how rarely this ought to happen, it's OK and arguably
                # slightly faster to do it here (ala conflicts above)
                if short:
                    prefix = 'X  '
                else:
                    prefix = ' '
                to_file.write("{} {}\n".format(prefix, nonexistent))
            if (new_is_working_tree and show_pending):
                show_pending_merges(new, to_file, short, verbose=verbose)
            if nonexistents:
                raise errors.PathsDoNotExist(nonexistents)
            for hook in hooks['post_status']:
                hook(StatusHookParams(
                    old, new, to_file, versioned, show_ids, short, verbose,
                    specific_files=specific_files))


def _get_sorted_revisions(tip_revision, revision_ids, parent_map):
    """Get an iterator which will return the revisions in merge sorted order.

    This will build up a list of all nodes, such that only nodes in the list
    are referenced. It then uses MergeSorter to return them in 'merge-sorted'
    order.

    :param revision_ids: A set of revision_ids
    :param parent_map: The parent information for each node. Revisions which
        are considered ghosts should not be present in the map.
    :return: iterator from MergeSorter.iter_topo_order()
    """
    # MergeSorter requires that all nodes be present in the graph, so get rid
    # of any references pointing outside of this graph.
    parent_graph = {}
    for revision_id in revision_ids:
        if revision_id not in parent_map:  # ghost
            parent_graph[revision_id] = []
        else:
            # Only include parents which are in this sub-graph
            parent_graph[revision_id] = [p for p in parent_map[revision_id]
                                         if p in revision_ids]
    sorter = tsort.MergeSorter(parent_graph, tip_revision)
    return sorter.iter_topo_order()


def show_pending_merges(new, to_file, short=False, verbose=False):
    """Write out a display of pending merges in a working tree."""
    parents = new.get_parent_ids()
    if len(parents) < 2:
        return

    term_width = osutils.terminal_width()
    if term_width is not None:
        # we need one extra space for terminals that wrap on last char
        term_width = term_width - 1
    if short:
        first_prefix = 'P   '
        sub_prefix = 'P.   '
    else:
        first_prefix = '  '
        sub_prefix = '    '

    def show_log_message(rev, prefix):
        if term_width is None:
            width = term_width
        else:
            width = term_width - len(prefix)
        log_message = log_formatter.log_string(None, rev, width, prefix=prefix)
        to_file.write(log_message + '\n')

    pending = parents[1:]
    branch = new.branch
    last_revision = parents[0]
    if not short:
        if verbose:
            to_file.write('pending merges:\n')
        else:
            to_file.write('pending merge tips:'
                          ' (use -v to see all merge revisions)\n')
    graph = branch.repository.get_graph()
    other_revisions = [last_revision]
    log_formatter = log.LineLogFormatter(to_file)
    for merge in pending:
        try:
            rev = branch.repository.get_revision(merge)
        except errors.NoSuchRevision:
            # If we are missing a revision, just print out the revision id
            to_file.write(first_prefix + '(ghost) ' +
                          merge.decode('utf-8') + '\n')
            other_revisions.append(merge)
            continue

        # Log the merge, as it gets a slightly different formatting
        show_log_message(rev, first_prefix)
        if not verbose:
            continue

        # Find all of the revisions in the merge source, which are not in the
        # last committed revision.
        merge_extra = graph.find_unique_ancestors(merge, other_revisions)
        other_revisions.append(merge)
        merge_extra.discard(_mod_revision.NULL_REVISION)

        # Get a handle to all of the revisions we will need
        revisions = dict(branch.repository.iter_revisions(merge_extra))

        # Display the revisions brought in by this merge.
        rev_id_iterator = _get_sorted_revisions(merge, merge_extra,
                                                branch.repository.get_parent_map(merge_extra))
        # Skip the first node
        num, first, depth, eom = next(rev_id_iterator)
        if first != merge:
            raise AssertionError('Somehow we misunderstood how'
                                 ' iter_topo_order works %s != %s' % (first, merge))
        for num, sub_merge, depth, eom in rev_id_iterator:
            rev = revisions[sub_merge]
            if rev is None:
                to_file.write(sub_prefix + '(ghost) ' +
                              sub_merge.decode('utf-8') + '\n')
                continue
            show_log_message(revisions[sub_merge], sub_prefix)


def _filter_nonexistent(orig_paths, old_tree, new_tree):
    """Convert orig_paths to two sorted lists and return them.

    The first is orig_paths paths minus the items in the second list,
    and the second list is paths that are not in either inventory or
    tree (they don't qualify if they exist in the tree's inventory, or
    if they exist in the tree but are not versioned.)

    If either of the two lists is empty, return it as an empty list.

    This can be used by operations such as brz status that can accept
    unknown or ignored files.
    """
    mutter("check paths: %r", orig_paths)
    if not orig_paths:
        return orig_paths, []
    s = old_tree.filter_unversioned_files(orig_paths)
    s = new_tree.filter_unversioned_files(s)
    nonexistent = [path for path in s if not new_tree.has_filename(path)]
    remaining = [path for path in orig_paths if path not in nonexistent]
    # Sorting the 'remaining' list doesn't have much effect in
    # practice, since the various status output sections will sort
    # their groups individually.  But for consistency of this
    # function's API, it's better to sort both than just 'nonexistent'.
    return sorted(remaining), sorted(nonexistent)


class StatusHooks(_mod_hooks.Hooks):
    """A dictionary mapping hook name to a list of callables for status hooks.

    e.g. ['post_status'] Is the list of items to be called when the
    status command has finished printing the status.
    """

    def __init__(self):
        """Create the default hooks.

        These are all empty initially, because by default nothing should get
        notified.
        """
        _mod_hooks.Hooks.__init__(self, "breezy.status", "hooks")
        self.add_hook(
            'post_status',
            "Called with argument StatusHookParams after Breezy has "
            "displayed the status. StatusHookParams has the attributes "
            "(old_tree, new_tree, to_file, versioned, show_ids, short, "
            "verbose). The last four arguments correspond to the command "
            "line options specified by the user for the status command. "
            "to_file is the output stream for writing.",
            (2, 3))
        self.add_hook(
            'pre_status',
            "Called with argument StatusHookParams before Breezy "
            "displays the status. StatusHookParams has the attributes "
            "(old_tree, new_tree, to_file, versioned, show_ids, short, "
            "verbose). The last four arguments correspond to the command "
            "line options specified by the user for the status command. "
            "to_file is the output stream for writing.",
            (2, 3))


class StatusHookParams:
    """Object holding parameters passed to post_status hooks.

    :ivar old_tree: Start tree (basis tree) for comparison.
    :ivar new_tree: Working tree.
    :ivar to_file: If set, write to this file.
    :ivar versioned: Show only versioned files.
    :ivar show_ids: Show internal object ids.
    :ivar short: Use short status indicators.
    :ivar verbose: Verbose flag.
    """

    def __init__(self, old_tree, new_tree, to_file, versioned, show_ids,
                 short, verbose, specific_files=None):
        """Create a group of post_status hook parameters.

        :param old_tree: Start tree (basis tree) for comparison.
        :param new_tree: Working tree.
        :param to_file: If set, write to this file.
        :param versioned: Show only versioned files.
        :param show_ids: Show internal object ids.
        :param short: Use short status indicators.
        :param verbose: Verbose flag.
        :param specific_files: If set, a list of filenames whose status should be
            shown.  It is an error to give a filename that is not in the
            working tree, or in the working inventory or in the basis inventory.
        """
        self.old_tree = old_tree
        self.new_tree = new_tree
        self.to_file = to_file
        self.versioned = versioned
        self.show_ids = show_ids
        self.short = short
        self.verbose = verbose
        self.specific_files = specific_files

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return "<{}({}, {}, {}, {}, {}, {}, {}, {})>".format(
            self.__class__.__name__, self.old_tree, self.new_tree,
            self.to_file, self.versioned, self.show_ids, self.short,
            self.verbose, self.specific_files)


def _show_shelve_summary(params):
    """post_status hook to display a summary of shelves.

    :param params: StatusHookParams.
    """
    # Don't show shelves if status of specific files is being shown, only if
    # no file arguments have been passed
    if params.specific_files:
        return
    get_shelf_manager = getattr(params.new_tree, 'get_shelf_manager', None)
    if get_shelf_manager is None:
        return
    try:
        manager = get_shelf_manager()
    except ShelvingUnsupported:
        mutter('shelving not supported by tree, not displaying shelves.')
    else:
        shelves = manager.active_shelves()
        if shelves:
            singular = '%d shelf exists. '
            plural = '%d shelves exist. '
            if len(shelves) == 1:
                fmt = singular
            else:
                fmt = plural
            params.to_file.write(fmt % len(shelves))
            params.to_file.write('See "brz shelve --list" for details.\n')


hooks = StatusHooks()


hooks.install_named_hook('post_status', _show_shelve_summary,
                         'brz status')



# Copyright (C) 2006 Canonical Ltd
# Authors:  Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the GitMemoryTree class."""

from ... import errors
from ...transport import NoSuchFile
from . import TestCaseWithTransport


class TestMemoryTree(TestCaseWithTransport):

    def make_branch(self, path, format='git'):
        return super().make_branch(path, format=format)

    def make_branch_and_tree(self, path, format='git'):
        return super().make_branch_and_tree(path, format=format)

    def test_create_on_branch(self):
        """Creating a mutable tree on a trivial branch works."""
        branch = self.make_branch('branch')
        tree = branch.create_memorytree()
        self.assertEqual(branch.controldir, tree.controldir)
        self.assertEqual(branch, tree.branch)
        self.assertEqual([], tree.get_parent_ids())

    def test_create_on_branch_with_content(self):
        """Creating a mutable tree on a non-trivial branch works."""
        wt = self.make_branch_and_tree('sometree')
        self.build_tree(['sometree/foo'])
        wt.add(['foo'])
        rev_id = wt.commit('first post')
        tree = wt.branch.create_memorytree()
        with tree.lock_read():
            self.assertEqual([rev_id], tree.get_parent_ids())
            self.assertEqual(b'contents of sometree/foo\n',
                             tree.get_file('foo').read())

    def test_lock_tree_write(self):
        """Check we can lock_tree_write and unlock MemoryTrees."""
        branch = self.make_branch('branch')
        tree = branch.create_memorytree()
        tree.lock_tree_write()
        tree.unlock()

    def test_lock_tree_write_after_read_fails(self):
        """Check that we error when trying to upgrade a read lock to write."""
        branch = self.make_branch('branch')
        tree = branch.create_memorytree()
        tree.lock_read()
        self.assertRaises(errors.ReadOnlyError, tree.lock_tree_write)
        tree.unlock()

    def test_lock_write(self):
        """Check we can lock_write and unlock MemoryTrees."""
        branch = self.make_branch('branch')
        tree = branch.create_memorytree()
        tree.lock_write()
        tree.unlock()

    def test_lock_write_after_read_fails(self):
        """Check that we error when trying to upgrade a read lock to write."""
        branch = self.make_branch('branch')
        tree = branch.create_memorytree()
        tree.lock_read()
        self.assertRaises(errors.ReadOnlyError, tree.lock_write)
        tree.unlock()

    def test_add_with_kind(self):
        branch = self.make_branch('branch')
        tree = branch.create_memorytree()
        tree.lock_write()
        tree.add(['', 'afile', 'adir'], ['directory', 'file', 'directory'])
        self.assertTrue(tree.is_versioned('afile'))
        self.assertFalse(tree.is_versioned('adir'))
        self.assertFalse(tree.has_filename('afile'))
        self.assertFalse(tree.has_filename('adir'))
        tree.unlock()

    def test_put_new_file(self):
        branch = self.make_branch('branch')
        tree = branch.create_memorytree()
        with tree.lock_write():
            tree.add(['', 'foo'], kinds=['directory', 'file'])
            tree.put_file_bytes_non_atomic('foo', b'barshoom')
            self.assertEqual(b'barshoom', tree.get_file('foo').read())

    def test_put_existing_file(self):
        branch = self.make_branch('branch')
        tree = branch.create_memorytree()
        with tree.lock_write():
            tree.add(['', 'foo'], kinds=['directory', 'file'])
            tree.put_file_bytes_non_atomic('foo', b'first-content')
            tree.put_file_bytes_non_atomic('foo', b'barshoom')
            self.assertEqual(b'barshoom', tree.get_file('foo').read())

    def test_add_in_subdir(self):
        branch = self.make_branch('branch')
        tree = branch.create_memorytree()
        with tree.lock_write():
            tree.add([''], ['directory'])
            tree.mkdir('adir')
            tree.put_file_bytes_non_atomic('adir/afile', b'barshoom')
            tree.add(['adir/afile'], ['file'])
            self.assertTrue(tree.is_versioned('adir/afile'))
            self.assertTrue(tree.is_versioned('adir'))

    def test_commit_trivial(self):
        """Smoke test for commit on a MemoryTree.

        Becamse of commits design and layering, if this works, all commit
        logic should work quite reliably.
        """
        branch = self.make_branch('branch')
        tree = branch.create_memorytree()
        with tree.lock_write():
            tree.add(['', 'foo'], kinds=['directory', 'file'])
            tree.put_file_bytes_non_atomic('foo', b'barshoom')
            revision_id = tree.commit('message baby')
            # the parents list for the tree should have changed.
            self.assertEqual([revision_id], tree.get_parent_ids())
        # and we should have a revision that is accessible outside the tree lock
        revtree = tree.branch.repository.revision_tree(revision_id)
        with revtree.lock_read():
            self.assertEqual(b'barshoom', revtree.get_file('foo').read())

    def test_unversion(self):
        """Some test for unversion of a memory tree."""
        branch = self.make_branch('branch')
        tree = branch.create_memorytree()
        with tree.lock_write():
            tree.add(['', 'foo'], kinds=['directory', 'file'])
            tree.unversion(['foo'])
            self.assertFalse(tree.is_versioned('foo'))
            self.assertFalse(tree.has_filename('foo'))

    def test_last_revision(self):
        """There should be a last revision method we can call."""
        tree = self.make_branch_and_memory_tree('branch')
        with tree.lock_write():
            tree.add('')
            rev_id = tree.commit('first post')
        self.assertEqual(rev_id, tree.last_revision())

    def test_rename_file(self):
        tree = self.make_branch_and_memory_tree('branch')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.add(['', 'foo'], ['directory', 'file'], ids=[b'root-id', b'foo-id'])
        tree.put_file_bytes_non_atomic('foo', b'content\n')
        tree.commit('one', rev_id=b'rev-one')
        tree.rename_one('foo', 'bar')
        self.assertEqual('bar', tree.id2path(b'foo-id'))
        self.assertEqual(b'content\n', tree._file_transport.get_bytes('bar'))
        self.assertRaises(NoSuchFile,
                          tree._file_transport.get_bytes, 'foo')
        tree.commit('two', rev_id=b'rev-two')
        self.assertEqual(b'content\n', tree._file_transport.get_bytes('bar'))
        self.assertRaises(NoSuchFile,
                          tree._file_transport.get_bytes, 'foo')

        rev_tree2 = tree.branch.repository.revision_tree(b'rev-two')
        self.assertEqual('bar', rev_tree2.id2path(b'foo-id'))
        self.assertEqual(b'content\n', rev_tree2.get_file_text('bar'))

    def test_rename_file_to_subdir(self):
        tree = self.make_branch_and_memory_tree('branch')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.add('')
        tree.mkdir('subdir', b'subdir-id')
        tree.add('foo', 'file', b'foo-id')
        tree.put_file_bytes_non_atomic('foo', b'content\n')
        tree.commit('one', rev_id=b'rev-one')

        tree.rename_one('foo', 'subdir/bar')
        self.assertEqual('subdir/bar', tree.id2path(b'foo-id'))
        self.assertEqual(b'content\n',
                         tree._file_transport.get_bytes('subdir/bar'))
        tree.commit('two', rev_id=b'rev-two')
        rev_tree2 = tree.branch.repository.revision_tree(b'rev-two')
        self.assertEqual('subdir/bar', rev_tree2.id2path(b'foo-id'))



# Copyright (C) 2005-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Messages and logging.

Messages are supplied by callers as a string-formatting template, plus values
to be inserted into it.  The actual %-formatting is deferred to the log
library so that it doesn't need to be done for messages that won't be emitted.

Messages are classified by severity levels: critical, error, warning, info,
and debug.

They can be sent to two places: stderr, and `$XDG_CACHE_HOME/breezy/brz.log`.
For purposes such as running the test suite, they can also be redirected away
from both of those two places to another location.

`brz.log` gets all messages, and full tracebacks for uncaught exceptions.
This trace file is always in UTF-8, regardless of the user's default encoding,
so that we can always rely on writing any message.

Output to stderr depends on the mode chosen by the user.  By default, messages
of info and above are sent out, which results in progress messages such as the
list of files processed by add and commit.  In debug mode, stderr gets debug messages too.

Errors that terminate an operation are generally passed back as exceptions;
others may be just emitted as messages.

Exceptions are reported in a brief form to stderr so as not to look scary.
BzrErrors are required to be able to format themselves into a properly
explanatory message.  This is not true for builtin exceptions such as
KeyError, which typically just str to "0".  They're printed in a different
form.
"""

# FIXME: Unfortunately it turns out that python's logging module
# is quite expensive, even when the message is not printed by any handlers.
# We should perhaps change back to just simply doing it here.
#
# On the other hand, as of 1.2 we generally only call the mutter() statement
# if (according to debug_flags) we actually intend to write it.  So the
# increased cost of logging.py is not so bad, and we could standardize on
# that.

import errno
import logging
import os
import sys
import time
from io import StringIO

import breezy

from .lazy_import import lazy_import

lazy_import(globals(), """
from breezy import (
    bedding,
    debug,
    osutils,
    ui,
    )
""")
from . import errors

# global verbosity for breezy; controls the log level for stderr; 0=normal; <0
# is quiet; >0 is verbose.
_verbosity_level = 0

# File-like object where mutter/debug output is currently sent.  Can be
# changed by _push_log_file etc.  This is directly manipulated by some
# external code; maybe there should be functions to do that more precisely
# than push/pop_log_file.
_trace_file = None

# Absolute path for brz.log.  Not changed even if the log/trace output is
# redirected elsewhere.  Used to show the location in --version.
_brz_log_filename = None

# The time the first message was written to the trace file, so that we can
# show relative times since startup.
_brz_log_start_time = breezy._start_time


# held in a global for quick reference
_brz_logger = logging.getLogger('brz')


def note(*args, **kwargs):
    """Output a note to the user.

    Takes the same parameters as logging.info.

    :return: None
    """
    # FIXME: clearing the ui and then going through the abstract logging
    # framework is whack; we should probably have a logging Handler that
    # deals with terminal output if needed.
    ui.ui_factory.clear_term()
    _brz_logger.info(*args, **kwargs)


def warning(*args, **kwargs):
    ui.ui_factory.clear_term()
    _brz_logger.warning(*args, **kwargs)


def show_error(*args, **kwargs):
    """Show an error message to the user.

    Don't use this for exceptions, use report_exception instead.
    """
    _brz_logger.error(*args, **kwargs)


class _Bytes(str):
    """Compat class for displaying bytes on Python 2."""

    def __repr__(self):
        return 'b' + str.__repr__(self)

    def __unicode__(self):
        return self.decode('ascii', 'replace')


def mutter(fmt, *args):
    if _trace_file is None:
        return
    # XXX: Don't check this every time; instead anyone who closes the file
    # ought to deregister it.  We can tolerate None.
    if (getattr(_trace_file, 'closed', None) is not None) and _trace_file.closed:
        return

    # Let format strings be specified as ascii bytes to help Python 2
    if isinstance(fmt, bytes):
        fmt = fmt.decode('ascii', 'replace')

    if args:
        out = fmt % args
    else:
        out = fmt
    now = time.time()
    out = '{:0.3f}  {}\n'.format(now - _brz_log_start_time, out)
    _trace_file.write(out.encode('utf-8'))
    # there's no explicit flushing; the file is typically line buffered.


def mutter_callsite(stacklevel, fmt, *args):
    """Perform a mutter of fmt and args, logging the call trace.

    :param stacklevel: The number of frames to show. None will show all
        frames.
    :param fmt: The format string to pass to mutter.
    :param args: A list of substitution variables.
    """
    import traceback
    outf = StringIO()
    if stacklevel is None:
        limit = None
    else:
        limit = stacklevel + 1
    traceback.print_stack(limit=limit, file=outf)
    formatted_lines = outf.getvalue().splitlines()
    formatted_stack = '\n'.join(formatted_lines[:-2])
    mutter(fmt + "\nCalled from:\n%s", *(args + (formatted_stack,)))


def _rollover_trace_maybe(trace_fname):
    import stat
    try:
        size = os.stat(trace_fname)[stat.ST_SIZE]
        if size <= 4 << 20:
            return
        old_fname = trace_fname + '.old'
        osutils.rename(trace_fname, old_fname)
    except OSError:
        return


def _get_brz_log_filename():
    """Return the brz log filename.

    :return: A path to the log file
    :raise EnvironmentError: If the cache directory could not be created
    """
    brz_log = os.environ.get('BRZ_LOG')
    if brz_log:
        return brz_log
    return os.path.join(bedding.cache_dir(), 'brz.log')


def _open_brz_log():
    """Open the brz.log trace file.

    If the log is more than a particular length, the old file is renamed to
    brz.log.old and a new file is started.  Otherwise, we append to the
    existing file.

    This sets the global _brz_log_filename.
    """
    global _brz_log_filename

    def _open_or_create_log_file(filename):
        """Open existing log file, or create with ownership and permissions

        It inherits the ownership and permissions (masked by umask) from
        the containing directory to cope better with being run under sudo
        with $HOME still set to the user's homedir.
        """
        flags = os.O_WRONLY | os.O_APPEND | osutils.O_TEXT
        while True:
            try:
                fd = os.open(filename, flags)
                break
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise
            try:
                fd = os.open(filename, flags | os.O_CREAT | os.O_EXCL, 0o666)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
            else:
                osutils.copy_ownership_from_path(filename)
                break
        return os.fdopen(fd, 'ab', 0)  # unbuffered

    try:
        _brz_log_filename = _get_brz_log_filename()
        _rollover_trace_maybe(_brz_log_filename)

        brz_log_file = _open_or_create_log_file(_brz_log_filename)
        brz_log_file.write(b'\n')
        if brz_log_file.tell() <= 2:
            brz_log_file.write(
                b"this is a debug log for diagnosing/reporting problems in brz\n")
            brz_log_file.write(
                b"you can delete or truncate this file, or include sections in\n")
            brz_log_file.write(
                b"bug reports to https://bugs.launchpad.net/brz/+filebug\n\n")

        return brz_log_file

    except OSError as e:
        # If we are failing to open the log, then most likely logging has not
        # been set up yet. So we just write to stderr rather than using
        # 'warning()'. If we using warning(), users get the unhelpful 'no
        # handlers registered for "brz"' when something goes wrong on the
        # server. (bug #503886)
        sys.stderr.write("failed to open trace file: {}\n".format(e))
    # TODO: What should happen if we fail to open the trace file?  Maybe the
    # objects should be pointed at /dev/null or the equivalent?  Currently
    # returns None which will cause failures later.
    return None


def enable_default_logging():
    """Configure default logging: messages to stderr and debug to brz.log

    This should only be called once per process.

    Non-command-line programs embedding breezy do not need to call this.  They
    can instead either pass a file to _push_log_file, or act directly on
    logging.getLogger("brz").

    Output can be redirected away by calling _push_log_file.

    :return: A memento from push_log_file for restoring the log state.
    """
    start_time = osutils.format_local_date(_brz_log_start_time,
                                           timezone='local')
    brz_log_file = _open_brz_log()
    if brz_log_file is not None:
        brz_log_file.write(start_time.encode('utf-8') + b'\n')
    memento = push_log_file(
        brz_log_file,
        r'[%(process)5d] %(asctime)s.%(msecs)03d %(levelname)s: %(message)s',
        r'%Y-%m-%d %H:%M:%S')
    # after hooking output into brz_log, we also need to attach a stderr
    # handler, writing only at level info and with encoding
    stderr_handler = logging.StreamHandler(stream=sys.stderr)
    logging.getLogger('brz').addHandler(stderr_handler)
    return memento


def push_log_file(to_file, log_format=None, date_format=None):
    """Intercept log and trace messages and send them to a file.

    :param to_file: A file-like object to which messages will be sent.

    :returns: A memento that should be passed to _pop_log_file to restore the
        previously active logging.
    """
    global _trace_file
    # make a new handler
    new_handler = EncodedStreamHandler(to_file, "utf-8", level=logging.DEBUG)
    if log_format is None:
        log_format = '%(levelname)8s  %(message)s'
    new_handler.setFormatter(logging.Formatter(log_format, date_format))
    # save and remove any existing log handlers
    brz_logger = logging.getLogger('brz')
    old_handlers = brz_logger.handlers[:]
    del brz_logger.handlers[:]
    # set that as the default logger
    brz_logger.addHandler(new_handler)
    brz_logger.setLevel(logging.DEBUG)
    # TODO: check if any changes are needed to the root logger
    #
    # TODO: also probably need to save and restore the level on brz_logger.
    # but maybe we can avoid setting the logger level altogether, and just set
    # the level on the handler?
    #
    # save the old trace file
    old_trace_file = _trace_file
    # send traces to the new one
    _trace_file = to_file
    return ('log_memento', old_handlers, new_handler, old_trace_file, to_file)


def pop_log_file(entry):
    """Undo changes to logging/tracing done by _push_log_file.

    This flushes, but does not close the trace file (so that anything that was
    in it is output.

    Takes the memento returned from _push_log_file."""
    (magic, old_handlers, new_handler, old_trace_file, new_trace_file) = entry
    global _trace_file
    _trace_file = old_trace_file
    brz_logger = logging.getLogger('brz')
    brz_logger.removeHandler(new_handler)
    # must be closed, otherwise logging will try to close it at exit, and the
    # file will likely already be closed underneath.
    new_handler.close()
    brz_logger.handlers = old_handlers
    if new_trace_file is not None:
        new_trace_file.flush()


def log_exception_quietly():
    """Log the last exception to the trace file only.

    Used for exceptions that occur internally and that may be
    interesting to developers but not to users.  For example,
    errors loading plugins.
    """
    import traceback
    mutter(traceback.format_exc())


def set_verbosity_level(level):
    """Set the verbosity level.

    :param level: -ve for quiet, 0 for normal, +ve for verbose
    """
    global _verbosity_level
    _verbosity_level = level
    _update_logging_level(level < 0)
    ui.ui_factory.be_quiet(level < 0)


def get_verbosity_level():
    """Get the verbosity level.

    See set_verbosity_level() for values.
    """
    return _verbosity_level


def be_quiet(quiet=True):
    if quiet:
        set_verbosity_level(-1)
    else:
        set_verbosity_level(0)


def _update_logging_level(quiet=True):
    """Hide INFO messages if quiet."""
    if quiet:
        _brz_logger.setLevel(logging.WARNING)
    else:
        _brz_logger.setLevel(logging.INFO)


def is_quiet():
    """Is the verbosity level negative?"""
    return _verbosity_level < 0


def is_verbose():
    """Is the verbosity level positive?"""
    return _verbosity_level > 0


def debug_memory(message='', short=True):
    """Write out a memory dump."""
    if sys.platform == 'win32':
        from breezy import win32utils
        win32utils.debug_memory_win32api(message=message, short=short)
    else:
        _debug_memory_proc(message=message, short=short)


_short_fields = ('VmPeak', 'VmSize', 'VmRSS')


def _debug_memory_proc(message='', short=True):
    try:
        status_file = open('/proc/%s/status' % os.getpid(), 'rb')
    except OSError:
        return
    try:
        status = status_file.read()
    finally:
        status_file.close()
    if message:
        note(message)
    for line in status.splitlines():
        if not short:
            note(line)
        else:
            for field in _short_fields:
                if line.startswith(field):
                    note(line)
                    break


def _dump_memory_usage(err_file):
    import tempfile
    try:
        try:
            fd, name = tempfile.mkstemp(prefix="brz_memdump", suffix=".json")
            dump_file = os.fdopen(fd, 'w')
            from meliae import scanner
            scanner.dump_gc_objects(dump_file)
            err_file.write("Memory dumped to %s\n" % name)
        except ImportError:
            err_file.write("Dumping memory requires meliae module.\n")
            log_exception_quietly()
        except BaseException:
            err_file.write("Exception while dumping memory.\n")
            log_exception_quietly()
    finally:
        if dump_file is not None:
            dump_file.close()
        elif fd is not None:
            os.close(fd)


def _qualified_exception_name(eclass, unqualified_breezy_errors=False):
    """Give name of error class including module for non-builtin exceptions

    If `unqualified_breezy_errors` is True, errors specific to breezy will
    also omit the module prefix.
    """
    class_name = eclass.__name__
    module_name = eclass.__module__
    if module_name in ("builtins", "exceptions", "__main__") or (
            unqualified_breezy_errors and module_name == "breezy.errors"):
        return class_name
    return "{}.{}".format(module_name, class_name)


def report_exception(exc_info, err_file):
    """Report an exception to err_file (typically stderr) and to brz.log.

    This will show either a full traceback or a short message as appropriate.

    :return: The appropriate exit code for this error.
    """
    # Log the full traceback to brz.log
    log_exception_quietly()
    if 'error' in debug.debug_flags:
        print_exception(exc_info, err_file)
        return errors.EXIT_ERROR
    exc_type, exc_object, exc_tb = exc_info
    if isinstance(exc_object, KeyboardInterrupt):
        err_file.write("brz: interrupted\n")
        return errors.EXIT_ERROR
    elif isinstance(exc_object, MemoryError):
        err_file.write("brz: out of memory\n")
        if 'mem_dump' in debug.debug_flags:
            _dump_memory_usage(err_file)
        else:
            err_file.write("Use -Dmem_dump to dump memory to a file.\n")
        return errors.EXIT_ERROR
    elif isinstance(exc_object, ImportError) \
            and str(exc_object).startswith("No module named "):
        report_user_error(
            exc_info, err_file,
            'You may need to install this Python library separately.')
        return errors.EXIT_ERROR
    elif not getattr(exc_object, 'internal_error', True):
        report_user_error(exc_info, err_file)
        return errors.EXIT_ERROR
    elif isinstance(exc_object, EnvironmentError):
        if getattr(exc_object, 'errno', None) == errno.EPIPE:
            err_file.write("brz: broken pipe\n")
            return errors.EXIT_ERROR
        # Might be nice to catch all of these and show them as something more
        # specific, but there are too many cases at the moment.
        report_user_error(exc_info, err_file)
        return errors.EXIT_ERROR
    else:
        report_bug(exc_info, err_file)
        return errors.EXIT_INTERNAL_ERROR


def print_exception(exc_info, err_file):
    import traceback
    exc_type, exc_object, exc_tb = exc_info
    err_file.write("brz: ERROR: {}: {}\n".format(
        _qualified_exception_name(exc_type), exc_object))
    err_file.write('\n')
    traceback.print_exception(exc_type, exc_object, exc_tb, file=err_file)


# TODO: Should these be specially encoding the output?
def report_user_error(exc_info, err_file, advice=None):
    """Report to err_file an error that's not an internal error.

    These don't get a traceback unless -Derror was given.

    :param exc_info: 3-tuple from sys.exc_info()
    :param advice: Extra advice to the user to be printed following the
        exception.
    """
    err_file.write("brz: ERROR: {}\n".format(str(exc_info[1])))
    if advice:
        err_file.write("%s\n" % advice)


def report_bug(exc_info, err_file):
    """Report an exception that probably indicates a bug in brz"""
    from breezy.crash import report_bug
    report_bug(exc_info, err_file)


def _flush_stdout_stderr():
    # called from the breezy library finalizer returned by breezy.initialize()
    try:
        sys.stdout.flush()
        sys.stderr.flush()
    except ValueError:
        # On Windows, I get ValueError calling stdout.flush() on a closed
        # handle
        pass
    except OSError as e:
        import errno
        if e.errno in [errno.EINVAL, errno.EPIPE]:
            pass
        else:
            raise


def _flush_trace():
    # called from the breezy library finalizer returned by breezy.initialize()
    global _trace_file
    if _trace_file:
        _trace_file.flush()


class EncodedStreamHandler(logging.Handler):
    """Robustly write logging events to a stream using the specified encoding

    Messages are expected to be formatted to unicode, but UTF-8 byte strings
    are also accepted. An error during formatting or a str message in another
    encoding will be quitely noted as an error in the Bazaar log file.

    The stream is not closed so sys.stdout or sys.stderr may be passed.
    """

    def __init__(self, stream, encoding=None, errors='strict', level=0):
        logging.Handler.__init__(self, level)
        self.stream = stream
        if encoding is None:
            encoding = getattr(stream, "encoding", "ascii")
        self.encoding = encoding
        self.errors = errors

    def flush(self):
        flush = getattr(self.stream, "flush", None)
        if flush is not None:
            flush()

    def emit(self, record):
        try:
            if not isinstance(record.msg, str):
                msg = record.msg.decode("utf-8")
                record.msg = msg
            line = self.format(record)
            if not isinstance(line, str):
                line = line.decode("utf-8")
            self.stream.write(line.encode(self.encoding, self.errors) + b"\n")
        except Exception:
            log_exception_quietly()
            # Try saving the details that would have been logged in some form
            msg = args = "<Unformattable>"
            try:
                msg = repr(record.msg)
                args = repr(record.args)
            except Exception:
                pass
            # Using mutter() bypasses the logging module and writes directly
            # to the file so there's no danger of getting into a loop here.
            mutter("Logging record unformattable: %s %% %s", msg, args)


class Config:
    """Configuration of message tracing in breezy.

    This implements the context manager protocol and should manage any global
    variables still used. The default config used is DefaultConfig, but
    embedded uses of breezy may wish to use a custom manager.
    """

    def __enter__(self):
        return self  # This is bound to the 'as' clause in a with statement.

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False  # propogate exceptions.


class DefaultConfig(Config):
    """A default configuration for tracing of messages in breezy.

    This implements the context manager protocol.
    """

    def __enter__(self):
        self._original_filename = _brz_log_filename
        self._original_state = enable_default_logging()
        return self  # This is bound to the 'as' clause in a with statement.

    def __exit__(self, exc_type, exc_val, exc_tb):
        pop_log_file(self._original_state)
        global _brz_log_filename
        _brz_log_filename = self._original_filename
        return False  # propogate exceptions.



# Copyright (C) 2006, 2007, 2008, 2013 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Tests for the wedav plugin."""

import stat
from http.client import parse_headers
from io import StringIO

from breezy import errors, tests
from breezy.plugins.webdav import webdav
from breezy.tests import http_server


def _get_list_dir_apache2_depth_1_prop():
    return """<?xml version="1.0" encoding="utf-8"?>
<D:multistatus xmlns:D="DAV:" xmlns:ns0="DAV:">
    <D:response>
        <D:href>/19016477731212686926.835527/</D:href>
        <D:propstat>
            <D:prop>
            </D:prop>
            <D:status>HTTP/1.1 200 OK</D:status>
        </D:propstat>
    </D:response>
    <D:response>
        <D:href>/19016477731212686926.835527/a</D:href>
        <D:propstat>
            <D:prop>
            </D:prop>
            <D:status>HTTP/1.1 200 OK</D:status>
        </D:propstat>
    </D:response>
    <D:response>
        <D:href>/19016477731212686926.835527/b</D:href>
        <D:propstat>
            <D:prop>
            </D:prop>
            <D:status>HTTP/1.1 200 OK</D:status>
        </D:propstat>
    </D:response>
    <D:response>
        <D:href>/19016477731212686926.835527/c/</D:href>
        <D:propstat>
            <D:prop>
            </D:prop>
            <D:status>HTTP/1.1 200 OK</D:status>
        </D:propstat>
    </D:response>
</D:multistatus>"""


def _get_list_dir_apache2_depth_1_allprop():
    return """<?xml version="1.0" encoding="utf-8"?>
<D:multistatus xmlns:D="DAV:" xmlns:ns0="DAV:">
    <D:response xmlns:lp1="DAV:" xmlns:lp2="http://apache.org/dav/props/">
        <D:href>/</D:href>
        <D:propstat>
            <D:prop>
                <lp1:resourcetype><D:collection/></lp1:resourcetype>
                <lp1:creationdate>2008-06-08T10:50:38Z</lp1:creationdate>
                <lp1:getlastmodified>Sun, 08 Jun 2008 10:50:38 GMT</lp1:getlastmodified>
                <lp1:getetag>"da7f5a-cc-7722db80"</lp1:getetag>
                <D:supportedlock>
                    <D:lockentry>
                        <D:lockscope><D:exclusive/></D:lockscope>
                        <D:locktype><D:write/></D:locktype>
                    </D:lockentry>
                    <D:lockentry>
                        <D:lockscope><D:shared/></D:lockscope>
                        <D:locktype><D:write/></D:locktype>
                    </D:lockentry>
                </D:supportedlock>
                <D:lockdiscovery/>
            </D:prop>
            <D:status>HTTP/1.1 200 OK</D:status>
        </D:propstat>
    </D:response>
    <D:response xmlns:lp1="DAV:" xmlns:lp2="http://apache.org/dav/props/">
        <D:href>/executable</D:href>
        <D:propstat>
            <D:prop>
                <lp1:resourcetype/>
                <lp1:creationdate>2008-06-08T09:50:15Z</lp1:creationdate>
                <lp1:getcontentlength>14</lp1:getcontentlength>
                <lp1:getlastmodified>Sun, 08 Jun 2008 09:50:11 GMT</lp1:getlastmodified>
                <lp1:getetag>"da9f81-0-9ef33ac0"</lp1:getetag>
                <lp2:executable>T</lp2:executable>
                <D:supportedlock>
                    <D:lockentry>
                        <D:lockscope><D:exclusive/></D:lockscope>
                        <D:locktype><D:write/></D:locktype>
                    </D:lockentry>
                    <D:lockentry>
                        <D:lockscope><D:shared/></D:lockscope>
                        <D:locktype><D:write/></D:locktype>
                    </D:lockentry>
                </D:supportedlock>
                <D:lockdiscovery/>
            </D:prop>
            <D:status>HTTP/1.1 200 OK</D:status>
        </D:propstat>
    </D:response>
    <D:response xmlns:lp1="DAV:" xmlns:lp2="http://apache.org/dav/props/">
        <D:href>/read-only</D:href>
        <D:propstat>
            <D:prop>
                <lp1:resourcetype/>
                <lp1:creationdate>2008-06-08T09:50:11Z</lp1:creationdate>
                <lp1:getcontentlength>42</lp1:getcontentlength>
                <lp1:getlastmodified>Sun, 08 Jun 2008 09:50:11 GMT</lp1:getlastmodified>
                <lp1:getetag>"da9f80-0-9ef33ac0"</lp1:getetag>
                <lp2:executable>F</lp2:executable>
                <D:supportedlock>
                    <D:lockentry>
                        <D:lockscope><D:exclusive/></D:lockscope>
                        <D:locktype><D:write/></D:locktype>
                    </D:lockentry>
                    <D:lockentry>
                        <D:lockscope><D:shared/></D:lockscope>
                        <D:locktype><D:write/></D:locktype>
                    </D:lockentry>
                </D:supportedlock>
                <D:lockdiscovery/>
            </D:prop>
            <D:status>HTTP/1.1 200 OK</D:status>
        </D:propstat>
    </D:response>
    <D:response xmlns:lp1="DAV:" xmlns:lp2="http://apache.org/dav/props/">
        <D:href>/titi</D:href>
        <D:propstat>
            <D:prop>
                <lp1:resourcetype/>
                <lp1:creationdate>2008-06-08T09:49:53Z</lp1:creationdate>
                <lp1:getcontentlength>6</lp1:getcontentlength>
                <lp1:getlastmodified>Sun, 08 Jun 2008 09:49:53 GMT</lp1:getlastmodified>
                <lp1:getetag>"da8cbc-6-9de09240"</lp1:getetag>
                <lp2:executable>F</lp2:executable>
                <D:supportedlock>
                    <D:lockentry>
                        <D:lockscope><D:exclusive/></D:lockscope>
                        <D:locktype><D:write/></D:locktype>
                    </D:lockentry>
                    <D:lockentry>
                        <D:lockscope><D:shared/></D:lockscope>
                        <D:locktype><D:write/></D:locktype>
                    </D:lockentry>
                </D:supportedlock>
                <D:lockdiscovery/>
            </D:prop>
            <D:status>HTTP/1.1 200 OK</D:status>
        </D:propstat>
    </D:response>
    <D:response xmlns:lp1="DAV:" xmlns:lp2="http://apache.org/dav/props/">
        <D:href>/toto/</D:href>
        <D:propstat>
            <D:prop>
                <lp1:resourcetype><D:collection/></lp1:resourcetype>
                <lp1:creationdate>2008-06-06T08:07:07Z</lp1:creationdate>
                <lp1:getlastmodified>Fri, 06 Jun 2008 08:07:07 GMT</lp1:getlastmodified>
                <lp1:getetag>"da8cb9-44-f2ac20c0"</lp1:getetag>
                <D:supportedlock>
                    <D:lockentry>
                        <D:lockscope><D:exclusive/></D:lockscope>
                        <D:locktype><D:write/></D:locktype>
                    </D:lockentry>
                    <D:lockentry>
                        <D:lockscope><D:shared/></D:lockscope>
                        <D:locktype><D:write/></D:locktype>
                    </D:lockentry>
                </D:supportedlock>
                <D:lockdiscovery/>
            </D:prop>
            <D:status>HTTP/1.1 200 OK</D:status>
        </D:propstat>
    </D:response>
</D:multistatus>
"""


class TestDavSaxParser(tests.TestCase):

    def _extract_dir_content_from_str(self, str):
        return webdav._extract_dir_content(
            'http://localhost/blah', StringIO(str))

    def _extract_stat_from_str(self, str):
        return webdav._extract_stat_info(
            'http://localhost/blah', StringIO(str))

    def test_unkown_format_response(self):
        # Valid but unrelated xml
        example = """<document/>"""
        self.assertRaises(errors.InvalidHttpResponse,
                          self._extract_dir_content_from_str, example)

    def test_list_dir_malformed_response(self):
        # Invalid xml, neither multistatus nor response are properly closed
        example = """<?xml version="1.0" encoding="utf-8"?>
<D:multistatus xmlns:D="DAV:" xmlns:ns0="urn:uuid:c2f41010-65b3-11d1-a29f-00aa00c14882/">
<D:response>
<D:href>http://localhost/</D:href>"""
        self.assertRaises(errors.InvalidHttpResponse,
                          self._extract_dir_content_from_str, example)

    def test_list_dir_incomplete_format_response(self):
        # The information we need is not present
        example = """<?xml version="1.0" encoding="utf-8"?>
<D:multistatus xmlns:D="DAV:" xmlns:ns0="urn:uuid:c2f41010-65b3-11d1-a29f-00aa00c14882/">
<D:response>
<D:href>http://localhost/</D:href>
</D:response>
<D:response>
<D:href>http://localhost/titi</D:href>
</D:response>
<D:href>http://localhost/toto</D:href>
</D:multistatus>"""
        self.assertRaises(errors.NotADirectory,
                         self._extract_dir_content_from_str, example)

    def test_list_dir_apache2_example(self):
        example = _get_list_dir_apache2_depth_1_prop()
        self.assertRaises(errors.NotADirectory,
                         self._extract_dir_content_from_str, example)

    def test_list_dir_lighttpd_example(self):
        example = """<?xml version="1.0" encoding="utf-8"?>
<D:multistatus xmlns:D="DAV:" xmlns:ns0="urn:uuid:c2f41010-65b3-11d1-a29f-00aa00c14882/">
<D:response>
<D:href>http://localhost/</D:href>
</D:response>
<D:response>
<D:href>http://localhost/titi</D:href>
</D:response>
<D:response>
<D:href>http://localhost/toto</D:href>
</D:response>
</D:multistatus>"""
        self.assertRaises(errors.NotADirectory,
                         self._extract_dir_content_from_str, example)

    def test_list_dir_apache2_dir_depth_1_example(self):
        example = _get_list_dir_apache2_depth_1_allprop()
        self.assertEqual([('executable', False, 14, True),
                           ('read-only', False, 42, False),
                           ('titi', False, 6, False),
                           ('toto', True, -1, False)],
                          self._extract_dir_content_from_str(example))

    def test_stat_malformed_response(self):
        # Invalid xml, neither multistatus nor response are properly closed
        example = """<?xml version="1.0" encoding="utf-8"?>
<D:multistatus xmlns:D="DAV:" xmlns:ns0="urn:uuid:c2f41010-65b3-11d1-a29f-00aa00c14882/">
<D:response>
<D:href>http://localhost/</D:href>"""
        self.assertRaises(errors.InvalidHttpResponse,
                          self._extract_stat_from_str, example)

    def test_stat_incomplete_format_response(self):
        # The minimal information is present but doesn't conform to RFC 2518
        # (well, as I understand it since the reference servers disagree on
        # more than details).

        # The href below is not enclosed in a response element and is
        # therefore ignored.
        example = """<?xml version="1.0" encoding="utf-8"?>
<D:multistatus xmlns:D="DAV:" xmlns:ns0="urn:uuid:c2f41010-65b3-11d1-a29f-00aa00c14882/">
<D:href>http://localhost/toto</D:href>
</D:multistatus>"""
        self.assertRaises(errors.InvalidHttpResponse,
                          self._extract_stat_from_str, example)

    def test_stat_apache2_file_example(self):
        example = """<?xml version="1.0" encoding="utf-8"?>
<D:multistatus xmlns:D="DAV:" xmlns:ns0="DAV:">
<D:response xmlns:lp1="DAV:" xmlns:lp2="http://apache.org/dav/props/">
<D:href>/executable</D:href>
<D:propstat>
<D:prop>
<lp1:resourcetype/>
<lp1:creationdate>2008-06-08T09:50:15Z</lp1:creationdate>
<lp1:getcontentlength>12</lp1:getcontentlength>
<lp1:getlastmodified>Sun, 08 Jun 2008 09:50:11 GMT</lp1:getlastmodified>
<lp1:getetag>"da9f81-0-9ef33ac0"</lp1:getetag>
<lp2:executable>T</lp2:executable>
<D:supportedlock>
<D:lockentry>
<D:lockscope><D:exclusive/></D:lockscope>
<D:locktype><D:write/></D:locktype>
</D:lockentry>
<D:lockentry>
<D:lockscope><D:shared/></D:lockscope>
<D:locktype><D:write/></D:locktype>
</D:lockentry>
</D:supportedlock>
<D:lockdiscovery/>
</D:prop>
<D:status>HTTP/1.1 200 OK</D:status>
</D:propstat>
</D:response>
</D:multistatus>"""
        st = self._extract_stat_from_str(example)
        self.assertEqual(12, st.st_size)
        self.assertFalse(stat.S_ISDIR(st.st_mode))
        self.assertTrue(stat.S_ISREG(st.st_mode))
        self.assertTrue(st.st_mode & stat.S_IXUSR)

    def test_stat_apache2_dir_depth_1_example(self):
        example = _get_list_dir_apache2_depth_1_allprop()
        self.assertRaises(errors.InvalidHttpResponse,
                          self._extract_stat_from_str, example)

    def test_stat_apache2_dir_depth_0_example(self):
        example = """<?xml version="1.0" encoding="utf-8"?>
<D:multistatus xmlns:D="DAV:" xmlns:ns0="DAV:">
<D:response xmlns:lp1="DAV:" xmlns:lp2="http://apache.org/dav/props/">
<D:href>/</D:href>
<D:propstat>
<D:prop>
<lp1:resourcetype><D:collection/></lp1:resourcetype>
<lp1:creationdate>2008-06-08T10:50:38Z</lp1:creationdate>
<lp1:getlastmodified>Sun, 08 Jun 2008 10:50:38 GMT</lp1:getlastmodified>
<lp1:getetag>"da7f5a-cc-7722db80"</lp1:getetag>
<D:supportedlock>
<D:lockentry>
<D:lockscope><D:exclusive/></D:lockscope>
<D:locktype><D:write/></D:locktype>
</D:lockentry>
<D:lockentry>
<D:lockscope><D:shared/></D:lockscope>
<D:locktype><D:write/></D:locktype>
</D:lockentry>
</D:supportedlock>
<D:lockdiscovery/>
</D:prop>
<D:status>HTTP/1.1 200 OK</D:status>
</D:propstat>
</D:response>
</D:multistatus>
"""
        st = self._extract_stat_from_str(example)
        self.assertEqual(-1, st.st_size)
        self.assertTrue(stat.S_ISDIR(st.st_mode))
        self.assertTrue(st.st_mode & stat.S_IXUSR)


class CannedRequestHandler(http_server.TestingHTTPRequestHandler):
    """Request handler for a unique and pre-defined request.

    The only thing we care about here is that a request is emitted by the
    client and we send back a syntactically correct response.

    We expect to receive a *single* request nothing more (and we won't even
    check what request it is, we just read until an empty line).
    """

    def _handle_one_request(self):
        # The communication between the client and the server is achieved
        # through the server defined in the client test case.
        tcs = self.server.test_case_server
        requestline = self.rfile.readline()
        # Read headers
        parse_headers(self.rfile)
        if requestline.startswith(b'POST'):
            # The body should be a single line (or we don't know where it ends
            # and we don't want to issue a blocking read)
            self.rfile.readline()

        self.wfile.write(tcs.canned_response)


class HatterHttpServer(http_server.HttpServer):
    """A server giving all sort of crazy responses (like Alice's Hatter).

    This is used to test various error cases in the webdav client.
    """

    def __init__(self):
        super().__init__(CannedRequestHandler,
                                               protocol_version='HTTP/1.1')
        self.canned_response = None


class TestDAVErrors(tests.TestCase):

    def setUp(self):
        super().setUp()
        self._transport = webdav.HttpDavTransport
        self.server = HatterHttpServer()
        self.server.start_server()
        self.addCleanup(self.server.stop_server)

    def get_transport(self):
        t = self._transport(self.server.get_url())
        return t

    def test_delete_replies_202(self):
        """A bogus return code for delete raises an error."""
        self.server.canned_response = b'''HTTP/1.1 202 OK\r
Date: Tue, 10 Aug 2013 14:38:56 GMT\r
Server: Apache/42 (Wonderland)\r
'''
        t = self.get_transport()
        self.assertRaises(errors.InvalidHttpResponse, t.delete, 'whatever')



# Copyright (C) 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from ... import bedding, errors, osutils
from ...tests import TestCase, TestCaseWithTransport
from ...tests.features import ModuleAvailableFeature

launchpadlib_feature = ModuleAvailableFeature('launchpadlib')


class TestDependencyManagement(TestCase):
    """Tests for managing the dependency on launchpadlib."""

    _test_needs_features = [launchpadlib_feature]

    def setUp(self):
        super().setUp()
        from . import lp_api
        self.lp_api = lp_api

    def patch(self, obj, name, value):
        """Temporarily set the 'name' attribute of 'obj' to 'value'."""
        self.overrideAttr(obj, name, value)

    def test_get_launchpadlib_version(self):
        # parse_launchpadlib_version returns a tuple of a version number of
        # the style used by launchpadlib.
        version_info = self.lp_api.parse_launchpadlib_version('1.5.1')
        self.assertEqual((1, 5, 1), version_info)

    def test_supported_launchpadlib_version(self):
        # If the installed version of launchpadlib is greater than the minimum
        # required version of launchpadlib, check_launchpadlib_compatibility
        # doesn't raise an error.
        launchpadlib = launchpadlib_feature.module
        self.patch(launchpadlib, '__version__', '1.5.1')
        self.lp_api.MINIMUM_LAUNCHPADLIB_VERSION = (1, 5, 1)
        # Doesn't raise an exception.
        self.lp_api.check_launchpadlib_compatibility()

    def test_unsupported_launchpadlib_version(self):
        # If the installed version of launchpadlib is less than the minimum
        # required version of launchpadlib, check_launchpadlib_compatibility
        # raises an DependencyNotPresent error.
        launchpadlib = launchpadlib_feature.module
        self.patch(launchpadlib, '__version__', '1.5.0')
        self.lp_api.MINIMUM_LAUNCHPADLIB_VERSION = (1, 5, 1)
        self.assertRaises(
            errors.DependencyNotPresent,
            self.lp_api.check_launchpadlib_compatibility)


class TestCacheDirectory(TestCase):
    """Tests for get_cache_directory."""

    _test_needs_features = [launchpadlib_feature]

    def test_get_cache_directory(self):
        # get_cache_directory returns the path to a directory inside the
        # Breezy cache directory.
        from . import lp_api
        try:
            expected_path = osutils.pathjoin(bedding.cache_dir(), 'launchpad')
        except OSError:
            self.assertRaises(EnvironmentError, lp_api.get_cache_directory)
        else:
            self.assertEqual(expected_path, lp_api.get_cache_directory())



# Copyright (C) 2007-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for breezy.switch."""


import os

from breezy import branch, errors
from breezy import merge as _mod_merge
from breezy import switch, tests, workingtree


class TestSwitch(tests.TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self.lightweight = True

    @staticmethod
    def _master_if_present(branch):
        master = branch.get_master_branch()
        if master:
            return master
        else:
            return branch

    def _setup_tree(self):
        tree = self.make_branch_and_tree('branch-1')
        self.build_tree(['branch-1/file-1'])
        tree.add('file-1')
        tree.commit('rev1')
        return tree

    def _setup_uncommitted(self, same_revision=False):
        tree = self._setup_tree()
        to_branch = tree.controldir.sprout('branch-2').open_branch()
        self.build_tree(['branch-1/file-2'])
        if not same_revision:
            tree.add('file-2')
            tree.remove('file-1')
            tree.commit('rev2')
        checkout = tree.branch.create_checkout('checkout',
                                               lightweight=self.lightweight)
        self.build_tree(['checkout/file-3'])
        checkout.add('file-3')
        return checkout, to_branch

    def test_switch_store_uncommitted(self):
        """Test switch updates tree and stores uncommitted changes."""
        checkout, to_branch = self._setup_uncommitted()
        self.assertPathDoesNotExist('checkout/file-1')
        self.assertPathExists('checkout/file-2')
        switch.switch(checkout.controldir, to_branch, store_uncommitted=True)
        self.assertPathExists('checkout/file-1')
        self.assertPathDoesNotExist('checkout/file-2')
        self.assertPathDoesNotExist('checkout/file-3')

    def test_switch_restore_uncommitted(self):
        """Test switch updates tree and restores uncommitted changes."""
        checkout, to_branch = self._setup_uncommitted()
        old_branch = self._master_if_present(checkout.branch)
        self.assertPathDoesNotExist('checkout/file-1')
        self.assertPathExists('checkout/file-2')
        self.assertPathExists('checkout/file-3')
        switch.switch(checkout.controldir, to_branch, store_uncommitted=True)
        checkout = workingtree.WorkingTree.open('checkout')
        switch.switch(checkout.controldir, old_branch, store_uncommitted=True)
        self.assertPathDoesNotExist('checkout/file-1')
        self.assertPathExists('checkout/file-2')
        self.assertPathExists('checkout/file-3')

    def test_switch_restore_uncommitted_same_revision(self):
        """Test switch updates tree and restores uncommitted changes."""
        checkout, to_branch = self._setup_uncommitted(same_revision=True)
        old_branch = self._master_if_present(checkout.branch)
        switch.switch(checkout.controldir, to_branch, store_uncommitted=True)
        checkout = workingtree.WorkingTree.open('checkout')
        switch.switch(checkout.controldir, old_branch, store_uncommitted=True)
        self.assertPathExists('checkout/file-3')

    def test_switch_updates(self):
        """Test switch updates tree and keeps uncommitted changes."""
        checkout, to_branch = self._setup_uncommitted()
        self.assertPathDoesNotExist('checkout/file-1')
        self.assertPathExists('checkout/file-2')
        switch.switch(checkout.controldir, to_branch)
        self.assertPathExists('checkout/file-1')
        self.assertPathDoesNotExist('checkout/file-2')
        self.assertPathExists('checkout/file-3')

    def test_switch_after_branch_moved(self):
        """Test switch after the branch is moved."""
        tree = self._setup_tree()
        checkout = tree.branch.create_checkout('checkout',
                                               lightweight=self.lightweight)
        self.build_tree(['branch-1/file-2'])
        tree.add('file-2')
        tree.remove('file-1')
        tree.commit('rev2')
        self.build_tree(['checkout/file-3'])
        checkout.add('file-3')
        # rename the branch on disk, the checkout object is now invalid.
        os.rename('branch-1', 'branch-2')
        to_branch = branch.Branch.open('branch-2')
        # Check fails without --force
        err = self.assertRaises(
            (errors.CommandError, errors.NotBranchError),
            switch.switch, checkout.controldir, to_branch)
        if isinstance(err, errors.CommandError):
            self.assertContainsRe(str(err),
                                  'Unable to connect to current master branch .*'
                                  'To switch anyway, use --force.')
        switch.switch(checkout.controldir, to_branch, force=True)
        self.assertPathDoesNotExist('checkout/file-1')
        self.assertPathExists('checkout/file-2')
        self.assertPathExists('checkout/file-3')

    def test_switch_when_pending_merges(self):
        """Test graceful failure if pending merges are outstanding."""
        # Create 2 branches and a checkout
        tree = self._setup_tree()
        tree2 = tree.controldir.sprout('branch-2').open_workingtree()
        checkout = tree.branch.create_checkout('checkout',
                                               lightweight=self.lightweight)
        # Change tree2 and merge it into the checkout without committing
        self.build_tree(['branch-2/file-2'])
        tree2.add('file-2')
        tree2.commit('rev2')
        checkout.merge_from_branch(tree2.branch)
        # Check the error reporting is as expected
        err = self.assertRaises(errors.CommandError,
                                switch.switch, checkout.controldir, tree2.branch)
        self.assertContainsRe(str(err),
                              "Pending merges must be committed or reverted before using switch")

    def test_switch_with_revision(self):
        """Test switch when a revision is given."""
        # Create a tree with 2 revisions
        tree = self.make_branch_and_tree('branch-1')
        self.build_tree(['branch-1/file-1'])
        tree.add('file-1')
        tree.commit(rev_id=b'rev1', message='rev1')
        self.build_tree(['branch-1/file-2'])
        tree.add('file-2')
        tree.commit(rev_id=b'rev2', message='rev2')
        # Check it out and switch to revision 1
        checkout = tree.branch.create_checkout('checkout',
                                               lightweight=self.lightweight)
        switch.switch(checkout.controldir, tree.branch, revision_id=b"rev1")
        self.assertPathExists('checkout/file-1')
        self.assertPathDoesNotExist('checkout/file-2')

    def test_switch_changing_root_id(self):
        tree = self._setup_tree()
        tree2 = self.make_branch_and_tree('tree-2')
        tree2.set_root_id(b'custom-root-id')
        self.build_tree(['tree-2/file-2'])
        tree2.add(['file-2'])
        tree2.commit('rev1b')
        checkout = tree.branch.create_checkout('checkout',
                                               lightweight=self.lightweight)
        switch.switch(checkout.controldir, tree2.branch)
        self.assertEqual(b'custom-root-id', tree2.path2id(''))

    def test_switch_configurable_file_merger(self):
        class DummyMerger(_mod_merge.ConfigurableFileMerger):
            name_prefix = 'file'

        _mod_merge.Merger.hooks.install_named_hook(
            'merge_file_content', DummyMerger,
            'test factory')
        foo = self.make_branch('foo')
        checkout = foo.create_checkout('checkout', lightweight=True)
        self.build_tree_contents([('checkout/file', b'a')])
        checkout.add('file')
        checkout.commit('a')
        bar = foo.controldir.sprout('bar').open_workingtree()
        self.build_tree_contents([('bar/file', b'b')])
        bar.commit('b')
        self.build_tree_contents([('checkout/file', b'c')])
        switch.switch(checkout.controldir, bar.branch)


class TestSwitchHeavyweight(TestSwitch):

    def setUp(self):
        super().setUp()
        self.lightweight = False

    def test_switch_with_local_commits(self):
        """Test switch complains about local commits unless --force given."""
        tree = self._setup_tree()
        to_branch = tree.controldir.sprout('branch-2').open_branch()
        self.build_tree(['branch-1/file-2'])
        tree.add('file-2')
        tree.remove('file-1')
        tree.commit('rev2')
        checkout = tree.branch.create_checkout('checkout')
        self.build_tree(['checkout/file-3'])
        checkout.add('file-3')
        checkout.commit(message='local only commit', local=True)
        self.build_tree(['checkout/file-4'])
        # Check the error reporting is as expected
        err = self.assertRaises(errors.CommandError,
                                switch.switch, checkout.controldir, to_branch)
        self.assertContainsRe(str(err),
                              'Cannot switch as local commits found in the checkout.')
        # Check all is ok when force is given
        self.assertPathDoesNotExist('checkout/file-1')
        self.assertPathExists('checkout/file-2')
        switch.switch(checkout.controldir, to_branch, force=True)
        self.assertPathExists('checkout/file-1')
        self.assertPathDoesNotExist('checkout/file-2')
        self.assertPathDoesNotExist('checkout/file-3')
        self.assertPathExists('checkout/file-4')
        # Check that the checkout is a true mirror of the bound branch
        self.assertEqual(to_branch.last_revision_info(),
                         checkout.branch.last_revision_info())



# Copyright (C) 2007-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from ..transport import Transport
# SFTPTransport is the only bundled transport that properly counts connections
# at the moment.
from . import test_sftp_transport


class TestCaseWithConnectionHookedTransport(test_sftp_transport.TestCaseWithSFTPServer):

    def setUp(self):
        super().setUp()
        self.reset_connections()

    def start_logging_connections(self):
        Transport.hooks.install_named_hook('post_connect',
                                           self.connections.append, None)

    def reset_connections(self):
        self.connections = []



# Copyright (C) 2006-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Report on version of breezy"""

import os
import platform
import sys

import breezy

from . import bedding, controldir, errors, osutils, trace


def show_version(show_config=True, show_copyright=True, to_file=None):
    if to_file is None:
        to_file = sys.stdout
    to_file.write("Breezy (brz) %s\n" % breezy.__version__)
    # is breezy itself in a branch?
    src_tree = _get_brz_source_tree()
    if src_tree:
        src_revision_id = src_tree.last_revision()
        to_file.write("  from brz checkout {}\n".format(src_tree.basedir))
        try:
            revno = src_tree.branch.revision_id_to_revno(src_revision_id)
        except errors.GhostRevisionsHaveNoRevno:
            pass
        else:
            to_file.write("    revision: {}\n".format(revno))
        to_file.write("    revid: {}\n".format(src_revision_id))
        to_file.write("    branch nick: {}\n".format(src_tree.branch.nick))

    to_file.write("  Python interpreter: ")
    # show path to python interpreter
    # (brz.exe use python interpreter from pythonXY.dll
    # but sys.executable point to brz.exe itself)
    # however, sys.frozen exists if running from brz.exe
    # see http://www.py2exe.org/index.cgi/Py2exeEnvironment
    if getattr(sys, 'frozen', None) is None:  # if not brz.exe
        to_file.write(sys.executable + ' ')
    else:
        # pythonXY.dll
        basedir = os.path.dirname(sys.executable)
        python_dll = "python%d%d.dll" % sys.version_info[:2]
        to_file.write(os.path.join(basedir, python_dll) + ' ')
    # and now version of python interpreter
    to_file.write(breezy._format_version_tuple(sys.version_info))
    to_file.write('\n')

    to_file.write("  Python standard library:" + ' ')
    to_file.write(os.path.dirname(os.__file__) + '\n')
    platform_str = platform.platform(aliased=1)
    if not isinstance(platform_str, str):
        platform_str = platform_str.decode('utf-8')
    to_file.write("  Platform: %s\n" % platform_str)
    to_file.write("  breezy: ")
    if len(breezy.__path__) > 1:
        # print repr, which is a good enough way of making it clear it's
        # more than one element (eg ['/foo/bar', '/foo/brz'])
        to_file.write(repr(breezy.__path__) + '\n')
    else:
        to_file.write(breezy.__path__[0] + '\n')
    if show_config:
        config_dir = osutils.normpath(
            bedding.config_dir())  # use native slashes
        if not isinstance(config_dir, str):
            config_dir = config_dir.decode(osutils.get_user_encoding())
        to_file.write("  Breezy configuration: {}\n".format(config_dir))
        to_file.write("  Breezy log file: ")
        to_file.write(trace._brz_log_filename + '\n')
    if show_copyright:
        to_file.write('\n')
        to_file.write(breezy.__copyright__ + '\n')
        to_file.write("https://www.breezy-vcs.org/\n")
        to_file.write('\n')
        to_file.write(
            "brz comes with ABSOLUTELY NO WARRANTY.  brz is free software, and\n")
        to_file.write(
            "you may use, modify and redistribute it under the terms of the GNU\n")
        to_file.write("General Public License version 2 or later.\n")
    to_file.write('\n')


def _get_brz_source_tree():
    """Return the WorkingTree for brz source, if any.

    If brz is not being run from its working tree, returns None.
    """
    try:
        control = controldir.ControlDir.open_containing(__file__)[0]
        return control.open_workingtree(recommend_upgrade=False)
    except (errors.NotBranchError, errors.UnknownFormatError,
            errors.NoWorkingTree):
        return None



# Copyright (C) 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Branch opening with URL-based restrictions."""

import threading

from . import errors, trace, urlutils
from .branch import Branch
from .controldir import ControlDir, ControlDirFormat
from .transport import do_catching_redirections, get_transport


class BadUrl(errors.BzrError):

    _fmt = "Tried to access a branch from bad URL %(url)s."


class BranchReferenceForbidden(errors.BzrError):

    _fmt = ("Trying to mirror a branch reference and the branch type "
            "does not allow references.")


class BranchLoopError(errors.BzrError):
    """Encountered a branch cycle.

    A URL may point to a branch reference or it may point to a stacked branch.
    In either case, it's possible for there to be a cycle in these references,
    and this exception is raised when we detect such a cycle.
    """

    _fmt = "Encountered a branch cycle"""


class BranchOpenPolicy:
    """Policy on how to open branches.

    In particular, a policy determines which branches are okay to open by
    checking their URLs and deciding whether or not to follow branch
    references.
    """

    def should_follow_references(self):
        """Whether we traverse references when mirroring.

        Subclasses must override this method.

        If we encounter a branch reference and this returns false, an error is
        raised.

        :returns: A boolean to indicate whether to follow a branch reference.
        """
        raise NotImplementedError(self.should_follow_references)

    def transform_fallback_location(self, branch, url):
        """Validate, maybe modify, 'url' to be used as a stacked-on location.

        :param branch:  The branch that is being opened.
        :param url: The URL that the branch provides for its stacked-on
            location.
        :return: (new_url, check) where 'new_url' is the URL of the branch to
            actually open and 'check' is true if 'new_url' needs to be
            validated by check_and_follow_branch_reference.
        """
        raise NotImplementedError(self.transform_fallback_location)

    def check_one_url(self, url):
        """Check a URL.

        Subclasses must override this method.

        :param url: The source URL to check.
        :raise BadUrl: subclasses are expected to raise this or a subclass
            when it finds a URL it deems to be unacceptable.
        """
        raise NotImplementedError(self.check_one_url)


class _BlacklistPolicy(BranchOpenPolicy):
    """Branch policy that forbids certain URLs.

    This doesn't cope with various alternative spellings of URLs,
    with e.g. url encoding. It's mostly useful for tests.
    """

    def __init__(self, should_follow_references, bad_urls=None):
        if bad_urls is None:
            bad_urls = set()
        self._bad_urls = bad_urls
        self._should_follow_references = should_follow_references

    def should_follow_references(self):
        return self._should_follow_references

    def check_one_url(self, url):
        if url in self._bad_urls:
            raise BadUrl(url)

    def transform_fallback_location(self, branch, url):
        """See `BranchOpenPolicy.transform_fallback_location`.

        This class is not used for testing our smarter stacking features so we
        just do the simplest thing: return the URL that would be used anyway
        and don't check it.
        """
        return urlutils.join(branch.base, url), False


class AcceptAnythingPolicy(_BlacklistPolicy):
    """Accept anything, to make testing easier."""

    def __init__(self):
        super().__init__(True, set())


class WhitelistPolicy(BranchOpenPolicy):
    """Branch policy that only allows certain URLs."""

    def __init__(self, should_follow_references, allowed_urls=None,
                 check=False):
        if allowed_urls is None:
            allowed_urls = []
        self.allowed_urls = {url.rstrip('/') for url in allowed_urls}
        self.check = check

    def should_follow_references(self):
        return self._should_follow_references

    def check_one_url(self, url):
        if url.rstrip('/') not in self.allowed_urls:
            raise BadUrl(url)

    def transform_fallback_location(self, branch, url):
        """See `BranchOpenPolicy.transform_fallback_location`.

        Here we return the URL that would be used anyway and optionally check
        it.
        """
        return urlutils.join(branch.base, url), self.check


class SingleSchemePolicy(BranchOpenPolicy):
    """Branch open policy that rejects URLs not on the given scheme."""

    def __init__(self, allowed_scheme):
        self.allowed_scheme = allowed_scheme

    def should_follow_references(self):
        return True

    def transform_fallback_location(self, branch, url):
        return urlutils.join(branch.base, url), True

    def check_one_url(self, url):
        """Check that `url` is okay to open."""
        if urlutils.URL.from_string(str(url)).scheme != self.allowed_scheme:
            raise BadUrl(url)


class BranchOpener:
    """Branch opener which uses a URL policy.

    All locations that are opened (stacked-on branches, references) are
    checked against a policy object.

    The policy object is expected to have the following methods:
    * check_one_url
    * should_follow_references
    * transform_fallback_location
    """

    _threading_data = threading.local()

    def __init__(self, policy, probers=None):
        """Create a new BranchOpener.

        :param policy: The opener policy to use.
        :param probers: Optional list of probers to allow.
            Defaults to local and remote bzr probers.
        """
        self.policy = policy
        self._seen_urls = set()
        if probers is None:
            probers = ControlDirFormat.all_probers()
        self.probers = probers

    @classmethod
    def install_hook(cls):
        """Install the ``transform_fallback_location`` hook.

        This is done at module import time, but transform_fallback_locationHook
        doesn't do anything unless the `_active_openers` threading.Local
        object has a 'opener' attribute in this thread.

        This is in a module-level function rather than performed at module
        level so that it can be called in setUp for testing `BranchOpener`
        as breezy.tests.TestCase.setUp clears hooks.
        """
        Branch.hooks.install_named_hook(
            'transform_fallback_location',
            cls.transform_fallback_locationHook,
            'BranchOpener.transform_fallback_locationHook')

    def check_and_follow_branch_reference(self, url):
        """Check URL (and possibly the referenced URL).

        This method checks that `url` passes the policy's `check_one_url`
        method, and if `url` refers to a branch reference, it checks whether
        references are allowed and whether the reference's URL passes muster
        also -- recursively, until a real branch is found.

        :param url: URL to check
        :raise BranchLoopError: If the branch references form a loop.
        :raise BranchReferenceForbidden: If this opener forbids branch
            references.
        """
        while True:
            if url in self._seen_urls:
                raise BranchLoopError()
            self._seen_urls.add(url)
            self.policy.check_one_url(url)
            next_url = self.follow_reference(url)
            if next_url is None:
                return url
            url = next_url
            if not self.policy.should_follow_references():
                raise BranchReferenceForbidden(url)

    @classmethod
    def transform_fallback_locationHook(cls, branch, url):
        """Installed as the 'transform_fallback_location' Branch hook.

        This method calls `transform_fallback_location` on the policy object
        and either returns the url it provides or passes it back to
        check_and_follow_branch_reference.
        """
        try:
            opener = getattr(cls._threading_data, "opener")
        except AttributeError:
            return url
        new_url, check = opener.policy.transform_fallback_location(branch, url)
        if check:
            return opener.check_and_follow_branch_reference(new_url)
        else:
            return new_url

    def run_with_transform_fallback_location_hook_installed(
            self, callable, *args, **kw):
        if (self.transform_fallback_locationHook not in
                Branch.hooks['transform_fallback_location']):
            raise AssertionError("hook not installed")
        self._threading_data.opener = self
        try:
            return callable(*args, **kw)
        finally:
            del self._threading_data.opener
            # We reset _seen_urls here to avoid multiple calls to open giving
            # spurious loop exceptions.
            self._seen_urls = set()

    def _open_dir(self, url):
        """Simple BzrDir.open clone that only uses specific probers.

        :param url: URL to open
        :return: ControlDir instance
        """
        def redirected(transport, e, redirection_notice):
            self.policy.check_one_url(e.target)
            redirected_transport = transport._redirected_to(
                e.source, e.target)
            if redirected_transport is None:
                raise errors.NotBranchError(e.source)
            trace.note(
                '%s is%s redirected to %s',
                transport.base, e.permanently, redirected_transport.base)
            return redirected_transport

        def find_format(transport):
            last_error = errors.NotBranchError(transport.base)
            for prober_kls in self.probers:
                prober = prober_kls()
                try:
                    return transport, prober.probe_transport(transport)
                except errors.NotBranchError as e:
                    last_error = e
            else:
                raise last_error
        transport = get_transport(url)
        transport, format = do_catching_redirections(
            find_format, transport, redirected)
        return format.open(transport)

    def follow_reference(self, url):
        """Get the branch-reference value at the specified url.

        This exists as a separate method only to be overriden in unit tests.
        """
        controldir = self._open_dir(url)
        return controldir.get_branch_reference()

    def open(self, url, ignore_fallbacks=False):
        """Open the Bazaar branch at url, first checking it.

        What is acceptable means is defined by the policy's `follow_reference`
        and `check_one_url` methods.
        """
        if not isinstance(url, str):
            raise TypeError

        url = self.check_and_follow_branch_reference(url)

        def open_branch(url, ignore_fallbacks):
            dir = self._open_dir(url)
            return dir.open_branch(ignore_fallbacks=ignore_fallbacks)
        return self.run_with_transform_fallback_location_hook_installed(
            open_branch, url, ignore_fallbacks)


def open_only_scheme(allowed_scheme, url):
    """Open the branch at `url`, only accessing URLs on `allowed_scheme`.

    :raises BadUrl: An attempt was made to open a URL that was not on
        `allowed_scheme`.
    """
    return BranchOpener(SingleSchemePolicy(allowed_scheme)).open(url)


BranchOpener.install_hook()



# Copyright (C) 2005-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the breezy ui."""

import time

from testtools.matchers import *

from .. import config, tests
from .. import ui as _mod_ui
from ..bzr import remote
from ..ui import text as _mod_ui_text
from . import fixtures, ui_testing
from .testui import ProgressRecordingUIFactory


class TestUIConfiguration(tests.TestCaseInTempDir):

    def test_output_encoding_configuration(self):
        enc = next(fixtures.generate_unicode_encodings())
        config.GlobalStack().set('output_encoding', enc)
        IO = ui_testing.BytesIOWithEncoding
        ui = _mod_ui.make_ui_for_terminal(IO(), IO(), IO())
        output = ui.make_output_stream()
        self.assertEqual(output.encoding, enc)


class TestTextUIFactory(tests.TestCase):

    def test_text_factory_confirm(self):
        # turns into reading a regular boolean
        with ui_testing.TestUIFactory('n\n') as ui:
            self.assertEqual(
                False,
                ui.confirm_action(
                    'Should %(thing)s pass?',
                    'breezy.tests.test_ui.confirmation',
                    {'thing': 'this'}))

    def test_text_factory_ascii_password(self):
        ui = ui_testing.TestUIFactory('secret\n')
        with ui.nested_progress_bar():
            self.assertEqual('secret',
                             self.apply_redirected(ui.stdin, ui.stdout,
                                                   ui.stderr,
                                                   ui.get_password))
            # ': ' is appended to prompt
            self.assertEqual(': ', ui.stderr.getvalue())
            self.assertEqual('', ui.stdout.readline())
            # stdin should be empty
            self.assertEqual('', ui.stdin.readline())

    def test_text_factory_unicode_password(self):
        """Test a unicode password."""
        ui = ui_testing.TextUIFactory('baz\u1234')
        password = ui.get_password(
            'Hello \u1234 %(user)s', user='some\u1234')
        self.assertEqual('baz\u1234', password)
        self.assertEqual('Hello \u1234 some\u1234: ', ui.stderr.getvalue())
        # stdin and stdout should be empty
        self.assertEqual('', ui.stdin.readline())
        self.assertEqual('', ui.stdout.getvalue())

    def test_text_ui_get_boolean(self):
        stdin_text = (
            "y\n"  # True
            "n\n"  # False
            " \n y \n"  # True
            " no \n"  # False
            "yes with garbage\nY\n"  # True
            "not an answer\nno\n"  # False
            "I'm sure!\nyes\n"  # True
            "NO\n"  # False
            "foo\n")
        with ui_testing.TextUIFactory(stdin_text) as factory:
            self.assertEqual(True, factory.get_boolean(""))
            self.assertEqual(False, factory.get_boolean(""))
            self.assertEqual(True, factory.get_boolean(""))
            self.assertEqual(False, factory.get_boolean(""))
            self.assertEqual(True, factory.get_boolean(""))
            self.assertEqual(False, factory.get_boolean(""))
            self.assertEqual(True, factory.get_boolean(""))
            self.assertEqual(False, factory.get_boolean(""))
            self.assertEqual("foo\n", factory.stdin.read())
            # stdin should be empty
            self.assertEqual('', factory.stdin.readline())
            # return false on EOF
            self.assertEqual(False, factory.get_boolean(""))

    def test_text_ui_choose_bad_parameters(self):
        with ui_testing.TextUIFactory("") as factory:
            # invalid default index
            self.assertRaises(ValueError, factory.choose, "", "&Yes\n&No", 3)
            # duplicated choice
            self.assertRaises(
                ValueError, factory.choose, "", "&choice\n&ChOiCe")
            # duplicated shortcut
            self.assertRaises(
                ValueError, factory.choose, "", "&choice1\nchoi&ce2")

    def test_text_ui_choose_prompt_explicit(self):
        # choices with explicit shortcuts
        with ui_testing.TextUIFactory("") as factory:
            factory.choose("prompt", "&yes\n&No\nmore &info")
            self.assertEqual(
                "prompt ([y]es, [N]o, more [i]nfo): \n",
                factory.stderr.getvalue())

    def test_text_ui_choose_prompt_automatic(self):
        # automatic shortcuts
        with ui_testing.TextUIFactory("") as factory:
            factory.choose("prompt", "yes\nNo\nmore info")
            self.assertEqual(
                "prompt ([y]es, [N]o, [m]ore info): \n",
                factory.stderr.getvalue())

    def test_text_ui_choose_return_values(self):
        def choose():
            return factory.choose("", "&Yes\n&No\nMaybe\nmore &info", 3)
        stdin_text = (
            "y\n"  # 0
            "n\n"  # 1
            " \n"  # default: 3
            " no \n"  # 1
            "b\na\nd \n"  # bad shortcuts, all ignored
            "yes with garbage\nY\n"  # 0
            "not an answer\nno\n"  # 1
            "info\nmore info\n"  # 3
            "Maybe\n"  # 2
            "foo\n")
        with ui_testing.TextUIFactory(stdin_text) as factory:
            self.assertEqual(0, choose())
            self.assertEqual(1, choose())
            self.assertEqual(3, choose())
            self.assertEqual(1, choose())
            self.assertEqual(0, choose())
            self.assertEqual(1, choose())
            self.assertEqual(3, choose())
            self.assertEqual(2, choose())
            self.assertEqual("foo\n", factory.stdin.read())
            # stdin should be empty
            self.assertEqual('', factory.stdin.readline())
            # return None on EOF
            self.assertEqual(None, choose())

    def test_text_ui_choose_no_default(self):
        stdin_text = (
            " \n"  # no default, invalid!
            " yes \n"  # 0
            "foo\n")
        with ui_testing.TextUIFactory(stdin_text) as factory:
            self.assertEqual(0, factory.choose("", "&Yes\n&No"))
            self.assertEqual("foo\n", factory.stdin.read())

    def test_text_ui_get_integer(self):
        stdin_text = (
            "1\n"
            "  -2  \n"
            "hmmm\nwhat else ?\nCome on\nok 42\n4.24\n42\n")
        with ui_testing.TextUIFactory(stdin_text) as factory:
            self.assertEqual(1, factory.get_integer(""))
            self.assertEqual(-2, factory.get_integer(""))
            self.assertEqual(42, factory.get_integer(""))

    def test_text_factory_prompt(self):
        # see <https://launchpad.net/bugs/365891>
        with ui_testing.TextUIFactory() as factory:
            factory.prompt('foo %2e')
            self.assertEqual('', factory.stdout.getvalue())
            self.assertEqual('foo %2e', factory.stderr.getvalue())

    def test_text_factory_prompts_and_clears(self):
        # a get_boolean call should clear the pb before prompting
        out = ui_testing.StringIOAsTTY()
        self.overrideEnv('TERM', 'xterm')
        factory = ui_testing.TextUIFactory("yada\ny\n", stdout=out, stderr=out)
        with factory:
            pb = factory.nested_progress_bar()
            pb._avail_width = lambda: 79
            pb.show_bar = False
            pb.show_spinner = False
            pb.show_count = False
            pb.update("foo", 0, 1)
            self.assertEqual(
                True,
                self.apply_redirected(
                    None, factory.stdout, factory.stdout, factory.get_boolean,
                    "what do you want"))
            output = out.getvalue()
            self.assertContainsRe(output,
                                  "| foo *\r\r  *\r*")
            self.assertContainsString(
                output,
                r"what do you want? ([y]es, [n]o): what do you want? "
                r"([y]es, [n]o): ")
            # stdin should have been totally consumed
            self.assertEqual('', factory.stdin.readline())

    def test_text_tick_after_update(self):
        ui_factory = ui_testing.TextUIFactory()
        with ui_factory.nested_progress_bar() as pb:
            pb.update('task', 0, 3)
            # Reset the clock, so that it actually tries to repaint itself
            ui_factory._progress_view._last_repaint = time.time() - 1.0
            pb.tick()

    def test_text_ui_getusername(self):
        ui = ui_testing.TextUIFactory('someuser\n\n')
        self.assertEqual('someuser',
                         ui.get_username('Hello %(host)s', host='some'))
        self.assertEqual('Hello some: ', ui.stderr.getvalue())
        self.assertEqual('', ui.stdout.getvalue())
        self.assertEqual('', ui.get_username("Gebruiker"))
        # stdin should be empty
        self.assertEqual('', ui.stdin.readline())

    def test_text_ui_getusername_unicode(self):
        ui = ui_testing.TextUIFactory('someuser\u1234')
        username = ui.get_username('Hello %(host)s', host='some\u1234')
        self.assertEqual("someuser\u1234", username)
        self.assertEqual("Hello some\u1234: ", ui.stderr.getvalue())
        self.assertEqual('', ui.stdout.getvalue())

    def test_quietness(self):
        self.overrideEnv('BRZ_PROGRESS_BAR', 'text')
        ui_factory = ui_testing.TextUIFactory(
            stderr=ui_testing.StringIOAsTTY())
        with ui_factory:
            self.assertIsInstance(ui_factory._progress_view,
                                  _mod_ui_text.TextProgressView)
            ui_factory.be_quiet(True)
            self.assertIsInstance(ui_factory._progress_view,
                                  _mod_ui_text.NullProgressView)

    def test_text_ui_show_user_warning(self):
        from ..bzr.groupcompress_repo import RepositoryFormat2a
        from ..bzr.knitpack_repo import RepositoryFormatKnitPack5
        ui = ui_testing.TextUIFactory()
        remote_fmt = remote.RemoteRepositoryFormat()
        remote_fmt._network_name = RepositoryFormatKnitPack5().network_name()
        ui.show_user_warning(
            'cross_format_fetch', from_format=RepositoryFormat2a(),
            to_format=remote_fmt)
        self.assertEqual('', ui.stdout.getvalue())
        self.assertContainsRe(
            ui.stderr.getvalue(),
            "^Doing on-the-fly conversion from RepositoryFormat2a\\(\\) to "
            "RemoteRepositoryFormat\\(_network_name="
            "b?'Bazaar RepositoryFormatKnitPack5 \\(bzr 1.6\\)\\\\n'\\)\\.\n"
            "This may take some time. Upgrade the repositories to "
            "the same format for better performance\\.\n$")
        # and now with it suppressed please
        ui = ui_testing.TextUIFactory()
        ui.suppressed_warnings.add('cross_format_fetch')
        ui.show_user_warning(
            'cross_format_fetch', from_format=RepositoryFormat2a(),
            to_format=remote_fmt)
        self.assertEqual('', ui.stdout.getvalue())
        self.assertEqual('', ui.stderr.getvalue())


class TestTextUIOutputStream(tests.TestCase):
    """Tests for output stream that synchronizes with progress bar."""

    def test_output_clears_terminal(self):
        clear_calls = []

        uif = ui_testing.TextUIFactory()
        uif.clear_term = lambda: clear_calls.append('clear')

        stream = _mod_ui_text.TextUIOutputStream(
            uif, uif.stdout, 'utf-8', 'strict')
        stream.write("Hello world!\n")
        stream.write("there's more...\n")
        stream.writelines(["1\n", "2\n", "3\n"])

        self.assertEqual(uif.stdout.getvalue(),
                         "Hello world!\n"
                         "there's more...\n"
                         "1\n2\n3\n")
        self.assertEqual(['clear', 'clear', 'clear'],
                         clear_calls)

        stream.flush()


class UITests(tests.TestCase):

    def test_progress_construction(self):
        """TextUIFactory constructs the right progress view.
        """
        FileStringIO = ui_testing.StringIOWithEncoding
        TTYStringIO = ui_testing.StringIOAsTTY
        for (file_class, term, pb, expected_pb_class) in (
                # on an xterm, either use them or not as the user requests,
                # otherwise default on
                (TTYStringIO, 'xterm', 'none', _mod_ui_text.NullProgressView),
                (TTYStringIO, 'xterm', 'text', _mod_ui_text.TextProgressView),
                (TTYStringIO, 'xterm', None, _mod_ui_text.TextProgressView),
                # on a dumb terminal, again if there's explicit configuration
                # do it, otherwise default off
                (TTYStringIO, 'dumb', 'none', _mod_ui_text.NullProgressView),
                (TTYStringIO, 'dumb', 'text', _mod_ui_text.TextProgressView),
                (TTYStringIO, 'dumb', None, _mod_ui_text.NullProgressView),
                # on a non-tty terminal, it's null regardless of $TERM
                (FileStringIO, 'xterm', None, _mod_ui_text.NullProgressView),
                (FileStringIO, 'dumb', None, _mod_ui_text.NullProgressView),
                # however, it can still be forced on
                (FileStringIO, 'dumb', 'text', _mod_ui_text.TextProgressView),
                ):
            self.overrideEnv('TERM', term)
            self.overrideEnv('BRZ_PROGRESS_BAR', pb)
            stdin = file_class('')
            stderr = file_class()
            stdout = file_class()
            uif = _mod_ui.make_ui_for_terminal(stdin, stdout, stderr)
            self.assertIsInstance(
                uif, _mod_ui_text.TextUIFactory,
                "TERM={} BRZ_PROGRESS_BAR={} uif={!r}".format(term, pb, uif))
            self.assertIsInstance(
                uif.make_progress_view(),
                expected_pb_class,
                "TERM={} BRZ_PROGRESS_BAR={} uif={!r}".format(term, pb, uif))

    def test_text_ui_non_terminal(self):
        """Even on non-ttys, make_ui_for_terminal gives a text ui."""
        stdin = stderr = stdout = ui_testing.StringIOWithEncoding()
        for term_type in ['dumb', None, 'xterm']:
            self.overrideEnv('TERM', term_type)
            uif = _mod_ui.make_ui_for_terminal(stdin, stdout, stderr)
            self.assertIsInstance(uif, _mod_ui_text.TextUIFactory,
                                  'TERM={!r}'.format(term_type))


class SilentUITests(tests.TestCase):

    def test_silent_factory_get_password(self):
        # A silent factory that can't do user interaction can't get a
        # password.  Possibly it should raise a more specific error but it
        # can't succeed.
        ui = _mod_ui.SilentUIFactory()
        stdout = ui_testing.StringIOWithEncoding()
        self.assertRaises(
            NotImplementedError,
            self.apply_redirected,
            None, stdout, stdout, ui.get_password)
        # and it didn't write anything out either
        self.assertEqual('', stdout.getvalue())

    def test_silent_ui_getbool(self):
        factory = _mod_ui.SilentUIFactory()
        stdout = ui_testing.StringIOWithEncoding()
        self.assertRaises(
            NotImplementedError,
            self.apply_redirected,
            None, stdout, stdout, factory.get_boolean, "foo")


class TestUIFactoryTests(tests.TestCase):

    def test_test_ui_factory_progress(self):
        # there's no output; we just want to make sure this doesn't crash -
        # see https://bugs.launchpad.net/bzr/+bug/408201
        ui = ui_testing.TestUIFactory()
        with ui.nested_progress_bar() as pb:
            pb.update('hello')
            pb.tick()


class CannedInputUIFactoryTests(tests.TestCase):

    def test_canned_input_get_input(self):
        uif = _mod_ui.CannedInputUIFactory([True, 'mbp', 'password', 42])
        self.assertEqual(True, uif.get_boolean('Extra cheese?'))
        self.assertEqual('mbp', uif.get_username('Enter your user name'))
        self.assertEqual('password',
                         uif.get_password('Password for %(host)s',
                                          host='example.com'))
        self.assertEqual(42, uif.get_integer('And all that jazz ?'))


class TestBoolFromString(tests.TestCase):

    def assertIsTrue(self, s, accepted_values=None):
        res = _mod_ui.bool_from_string(s, accepted_values=accepted_values)
        self.assertEqual(True, res)

    def assertIsFalse(self, s, accepted_values=None):
        res = _mod_ui.bool_from_string(s, accepted_values=accepted_values)
        self.assertEqual(False, res)

    def assertIsNone(self, s, accepted_values=None):
        res = _mod_ui.bool_from_string(s, accepted_values=accepted_values)
        self.assertIs(None, res)

    def test_know_valid_values(self):
        self.assertIsTrue('true')
        self.assertIsFalse('false')
        self.assertIsTrue('1')
        self.assertIsFalse('0')
        self.assertIsTrue('on')
        self.assertIsFalse('off')
        self.assertIsTrue('yes')
        self.assertIsFalse('no')
        self.assertIsTrue('y')
        self.assertIsFalse('n')
        # Also try some case variations
        self.assertIsTrue('True')
        self.assertIsFalse('False')
        self.assertIsTrue('On')
        self.assertIsFalse('Off')
        self.assertIsTrue('ON')
        self.assertIsFalse('OFF')
        self.assertIsTrue('oN')
        self.assertIsFalse('oFf')

    def test_invalid_values(self):
        self.assertIsNone(None)
        self.assertIsNone('doubt')
        self.assertIsNone('frue')
        self.assertIsNone('talse')
        self.assertIsNone('42')

    def test_provided_values(self):
        av = dict(y=True, n=False, yes=True, no=False)
        self.assertIsTrue('y', av)
        self.assertIsTrue('Y', av)
        self.assertIsTrue('Yes', av)
        self.assertIsFalse('n', av)
        self.assertIsFalse('N', av)
        self.assertIsFalse('No', av)
        self.assertIsNone('1', av)
        self.assertIsNone('0', av)
        self.assertIsNone('on', av)
        self.assertIsNone('off', av)


class TestConfirmationUserInterfacePolicy(tests.TestCase):

    def test_confirm_action_default(self):
        base_ui = _mod_ui.NoninteractiveUIFactory()
        for answer in [True, False]:
            self.assertEqual(
                _mod_ui.ConfirmationUserInterfacePolicy(base_ui, answer, {})
                .confirm_action("Do something?",
                                "breezy.tests.do_something", {}),
                answer)

    def test_confirm_action_specific(self):
        base_ui = _mod_ui.NoninteractiveUIFactory()
        for default_answer in [True, False]:
            for specific_answer in [True, False]:
                for conf_id in ['given_id', 'other_id']:
                    wrapper = _mod_ui.ConfirmationUserInterfacePolicy(
                        base_ui, default_answer,
                        dict(given_id=specific_answer))
                    result = wrapper.confirm_action(
                        "Do something?", conf_id, {})
                    if conf_id == 'given_id':
                        self.assertEqual(result, specific_answer)
                    else:
                        self.assertEqual(result, default_answer)

    def test_repr(self):
        base_ui = _mod_ui.NoninteractiveUIFactory()
        wrapper = _mod_ui.ConfirmationUserInterfacePolicy(
            base_ui, True, dict(a=2))
        self.assertThat(repr(wrapper),
                        Equals("ConfirmationUserInterfacePolicy("
                               "NoninteractiveUIFactory(), True, {'a': 2})"))


class TestProgressRecordingUI(tests.TestCase):
    """Test test-oriented UIFactory that records progress updates"""

    def test_nested_ignore_depth_beyond_one(self):
        # we only want to capture the first level out progress, not
        # want sub-components might do. So we have nested bars ignored.
        factory = ProgressRecordingUIFactory()
        pb1 = factory.nested_progress_bar()
        pb1.update('foo', 0, 1)
        pb2 = factory.nested_progress_bar()
        pb2.update('foo', 0, 1)
        pb2.finished()
        pb1.finished()
        self.assertEqual([("update", 0, 1, 'foo')], factory._calls)
